# LIGHTNING/ZAP FUNCTIONS
import json
import os
import random
import string
import urllib.parse
from hashlib import sha256
from pathlib import Path

import bech32
import dotenv
import requests
from Crypto.Cipher import AES
from Crypto.Util.Padding import pad
from bech32 import bech32_decode, convertbits, bech32_encode
from nostr_sdk import PublicKey, SecretKey, Event, EventBuilder, Tag, Keys, generate_shared_key, Kind, \
    Timestamp

from nostr_dvm.utils.nostr_utils import get_event_by_id, check_and_decrypt_own_tags, update_profile_lnaddress

# TODO tor connection to lnbits
# proxies = {
#    'http': 'socks5h://127.0.0.1:9050',
#    'https': 'socks5h://127.0.0.1:9050'
# }

proxies = {}


async def parse_zap_event_tags(zap_event, keys, name, client, config):
    zapped_event = None
    invoice_amount = 0
    anon = False
    message = ""
    sender = zap_event.author()
    for tag in zap_event.tags():
        if tag.as_vec()[0] == 'bolt11':
            invoice_amount = parse_amount_from_bolt11_invoice(tag.as_vec()[1])
        elif tag.as_vec()[0] == 'e':
            zapped_event = await get_event_by_id(tag.as_vec()[1], client=client, config=config)
            if zapped_event is not None:
                zapped_event = check_and_decrypt_own_tags(zapped_event, config)
        elif tag.as_vec()[0] == 'p':
            p_tag = tag.as_vec()[1]
        elif tag.as_vec()[0] == 'description':
            zap_request_event = Event.from_json(tag.as_vec()[1])
            sender = check_for_zapplepay(zap_request_event.author().to_hex(),
                                         zap_request_event.content())
            for z_tag in zap_request_event.tags():
                if z_tag.as_vec()[0] == 'anon':
                    if len(z_tag.as_vec()) > 1:
                        # print("[" + name + "] Private Zap received.")
                        decrypted_content = decrypt_private_zap_message(z_tag.as_vec()[1],
                                                                        keys.secret_key(),
                                                                        zap_request_event.author())
                        decrypted_private_event = Event.from_json(decrypted_content)
                        if decrypted_private_event.kind().as_u16() == 9733:
                            sender = decrypted_private_event.author().to_hex()
                            message = decrypted_private_event.content()
                            # if message != "":
                            #    print("Zap Message: " + message)
                    else:
                        anon = True
                        print(
                            "[" + name + "] Anonymous Zap received. Unlucky, I don't know from whom, and never will")

    return invoice_amount, zapped_event, sender, message, anon


def parse_amount_from_bolt11_invoice(bolt11_invoice: str) -> int:
    def get_index_of_first_letter(ip):
        index = 0
        for c in ip:
            if c.isalpha():
                return index
            else:
                index = index + 1
        return len(ip)

    remaining_invoice = bolt11_invoice[4:]
    index = get_index_of_first_letter(remaining_invoice)
    identifier = remaining_invoice[index]
    number_string = remaining_invoice[:index]
    number = float(number_string)
    if identifier == 'm':
        number = number * 100000000 * 0.001
    elif identifier == 'u':
        number = number * 100000000 * 0.000001
    elif identifier == 'n':
        number = number * 100000000 * 0.000000001
    elif identifier == 'p':
        number = number * 100000000 * 0.000000000001

    return int(number)


def create_bolt11_ln_bits(sats: int, config) -> (str, str):
    if config.LNBITS_URL == "":
        return None, None
    url = config.LNBITS_URL + "/api/v1/payments"
    data = {'out': False, 'amount': sats, 'memo': "Nostr-DVM " + config.NIP89.NAME}
    headers = {'X-API-Key': config.LNBITS_INVOICE_KEY, 'Content-Type': 'application/json', 'charset': 'UTF-8'}
    try:
        res = requests.post(url, json=data, headers=headers)
        obj = json.loads(res.text)
        if obj.get("payment_request") and obj.get("payment_hash"):
            return obj["payment_request"], obj["payment_hash"]  #
        else:
            print("LNBITS: " + res.text)
            return None, None
    except Exception as e:
        print("LNBITS: " + str(e))
        return None, None


def create_bolt11_lud16(lud16, amount):
    if lud16.startswith("LNURL") or lud16.startswith("lnurl"):
        url = decode_bech32(lud16)
        print(url)
    elif '@' in lud16:  # LNaddress
        url = 'https://' + str(lud16).split('@')[1] + '/.well-known/lnurlp/' + str(lud16).split('@')[0]
    else:  # No lud16 set or format invalid
        return None
    try:
        print(url)
        response = requests.get(url)
        ob = json.loads(response.content)
        callback = ob["callback"]
        response = requests.get(callback + "?amount=" + str(int(amount) * 1000))
        ob = json.loads(response.content)
        return ob["pr"]
    except Exception as e:
        print("LUD16: " + e)
        return None


def create_lnbits_account(name):
    if os.getenv("LNBITS_WALLET_ID") is None or os.getenv("LNBITS_WALLET_ID") == "":
        print("No admin id set, no wallet created.")
        return "", "", "", "", "failed"
    data = {
        'admin_id': os.getenv("LNBITS_WALLET_ID"),
        'wallet_name': name,
        'user_name': name,
    }
    try:
        json_object = json.dumps(data)
        url = os.getenv("LNBITS_HOST") + '/usermanager/api/v1/users'
        print(url)
        headers = {'X-API-Key': os.getenv("LNBITS_ADMIN_KEY"), 'Content-Type': 'application/json', 'charset': 'UTF-8'}
        r = requests.post(url, data=json_object, headers=headers, proxies=proxies)
        walletjson = json.loads(r.text)
        print(walletjson)
        if walletjson.get("wallets"):
            return walletjson['wallets'][0]['inkey'], walletjson['wallets'][0]['adminkey'], walletjson['wallets'][0][
                'id'], walletjson['id'], "success"
    except:
        print("error creating wallet")
        return "", "", "", "", "failed"


def check_bolt11_ln_bits_is_paid(payment_hash: str, config):
    url = config.LNBITS_URL + "/api/v1/payments/" + payment_hash
    headers = {'X-API-Key': config.LNBITS_INVOICE_KEY, 'Content-Type': 'application/json', 'charset': 'UTF-8'}
    try:
        res = requests.get(url, headers=headers, proxies=proxies)
        obj = json.loads(res.text)
        if obj.get("paid"):
            return obj["paid"]
        else:
            return False
    except Exception as e:
        return None


def pay_bolt11_ln_bits(bolt11: str, config):
    url = config.LNBITS_URL + "/api/v1/payments"
    data = {'out': True, 'bolt11': bolt11}
    headers = {'X-API-Key': config.LNBITS_ADMIN_KEY, 'Content-Type': 'application/json', 'charset': 'UTF-8'}
    try:
        res = requests.post(url, json=data, headers=headers)
        obj = json.loads(res.text)
        if obj.get("payment_hash"):
            return obj["payment_hash"]
        else:
            return "Error"
    except Exception as e:
        print("LNBITS: " + str(e))
        return "Error"


# DECRYPT ZAPS
def check_for_zapplepay(pubkey_hex: str, content: str):
    try:
        # Special case Zapplepay
        if (pubkey_hex == PublicKey.from_bech32("npub1wxl6njlcgygduct7jkgzrvyvd9fylj4pqvll6p32h59wyetm5fxqjchcan")
                .to_hex()):
            real_sender_bech32 = content.replace("From: nostr:", "")
            pubkey_hex = PublicKey.from_bech32(real_sender_bech32).to_hex()
        return pubkey_hex

    except Exception as e:
        print(e)
        return pubkey_hex


def enrypt_private_zap_message(message, privatekey, publickey):
    # Generate a random IV
    shared_secret = generate_shared_key(privatekey, publickey)
    iv = os.urandom(16)

    # Encrypt the message
    cipher = AES.new(bytearray(shared_secret), AES.MODE_CBC, bytearray(iv))
    utf8message = message.encode('utf-8')
    padded_message = pad(utf8message, AES.block_size)
    encrypted_msg = cipher.encrypt(padded_message)

    encrypted_msg_bech32 = bech32_encode("pzap", convertbits(encrypted_msg, 8, 5, True))
    iv_bech32 = bech32_encode("iv", convertbits(iv, 8, 5, True))
    return encrypted_msg_bech32 + "_" + iv_bech32


def decrypt_private_zap_message(msg: str, privkey: SecretKey, pubkey: PublicKey):
    shared_secret = generate_shared_key(privkey, pubkey)
    if len(shared_secret) != 16 and len(shared_secret) != 32:
        return "invalid shared secret size"
    parts = msg.split("_")
    if len(parts) != 2:
        return "invalid message format"
    try:
        _, encrypted_msg = bech32_decode(parts[0])
        encrypted_bytes = convertbits(encrypted_msg, 5, 8, False)
        _, iv = bech32_decode(parts[1])
        iv_bytes = convertbits(iv, 5, 8, False)
    except Exception as e:
        return e
    try:
        cipher = AES.new(bytearray(shared_secret), AES.MODE_CBC, bytearray(iv_bytes))
        decrypted_bytes = cipher.decrypt(bytearray(encrypted_bytes))
        plaintext = decrypted_bytes.decode("utf-8")
        decoded = plaintext.rsplit("}", 1)[0] + "}"  # weird symbols at the end
        return decoded
    except Exception as ex:
        return str(ex)


def decode_bech32(encoded_lnurl):
    # Decode the bech32 encoded string
    hrp, data = bech32.bech32_decode(encoded_lnurl)

    if hrp != 'lnurl':
        raise ValueError("Invalid human-readable part (hrp)")

    # Convert the data back from 5-bit words to 8-bit bytes
    decoded_bytes = bech32.convertbits(data, 5, 8, False)

    # Convert the bytes back to a string
    decoded_url = bytes(decoded_bytes).decode('utf-8')

    return decoded_url


def zaprequest(lud16: str, amount: int, content, zapped_event, zapped_user, keys, relay_list, zaptype="public"):
    print(lud16)
    print(str(amount))
    print(content)
    if lud16.startswith("LNURL") or lud16.startswith("lnurl"):
        url = decode_bech32(lud16)
    elif '@' in lud16:  # LNaddress
        url = 'https://' + str(lud16).split('@')[1] + '/.well-known/lnurlp/' + str(lud16).split('@')[0]
    else:  # No lud16 set or format invalid
        return None
    try:
        response = requests.get(url)
        ob = json.loads(response.content)
        callback = ob["callback"]
        print(ob["callback"])

        # encoded_lnurl = lnurl.encode(url)

        url_bytes = url.encode()
        encoded_lnurl = bech32.bech32_encode('lnurl', bech32.convertbits(url_bytes, 8, 5))

        amount_tag = Tag.parse(['amount', str(amount * 1000)])
        relays_tag = Tag.parse(['relays', str(relay_list)])
        lnurl_tag = Tag.parse(['lnurl', encoded_lnurl])
        if zapped_event is not None:
            p_tag = Tag.parse(['p', zapped_event.author().to_hex()])
            e_tag = Tag.parse(['e', zapped_event.id().to_hex()])
            tags = [amount_tag, relays_tag, p_tag, e_tag, lnurl_tag]
        else:
            p_tag = Tag.parse(['p', zapped_user.to_hex()])
            tags = [amount_tag, relays_tag, p_tag, lnurl_tag]

        if zaptype == "private":
            if zapped_event is not None:
                key_str = keys.secret_key().to_hex() + zapped_event.id().to_hex() + str(
                    zapped_event.created_at().as_secs())

            else:
                key_str = keys.secret_key().to_hex() + str(Timestamp.now().as_secs())

            encryption_key = sha256(key_str.encode('utf-8')).hexdigest()
            tags = [p_tag]
            if zapped_event is not None:
                tags.append(e_tag)
            zap_request = EventBuilder(Kind(9733), content,
                                       tags).to_event(keys).as_json()
            keys = Keys.parse(encryption_key)
            if zapped_event is not None:
                encrypted_content = enrypt_private_zap_message(zap_request, keys.secret_key(), zapped_event.author())
            else:
                encrypted_content = enrypt_private_zap_message(zap_request, keys.secret_key(), zapped_user)

            anon_tag = Tag.parse(['anon', encrypted_content])
            tags.append(anon_tag)
            content = ""

        zap_request = EventBuilder(Kind(9734), content,
                                   tags).to_event(keys).as_json()

        response = requests.get(callback + "?amount=" + str(int(amount) * 1000) + "&nostr=" + urllib.parse.quote_plus(
            zap_request) + "&lnurl=" + encoded_lnurl)
        ob = json.loads(response.content)
        return ob["pr"]

    except Exception as e:
        print("ZAP REQUEST: " + str(e))
        return None


def get_price_per_sat(currency):
    import requests

    url = "https://openapiv1.coinstats.app/coins/bitcoin"
    params = {"skip": 0, "limit": 1, "currency": currency}
    price_currency_per_sat = 0.0004
    if os.getenv("COINSTATSOPENAPI_KEY"):

        header = {'accept': 'application/json', 'X-API-KEY': os.getenv("COINSTATSOPENAPI_KEY")}
        try:
            response = requests.get(url, headers=header, params=params)
            response_json = response.json()

            bitcoin_price = response_json["price"]
            price_currency_per_sat = bitcoin_price / 100000000.0
        except:
            price_currency_per_sat = 0.0004

    return price_currency_per_sat


def randomword(length):
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for i in range(length))


def make_ln_address_nostdress(identifier, npub, pin, nostdressdomain, newname=" ", currentname=" "):
    if newname == " ":
        newname = identifier

    print(os.getenv("LNBITS_INVOICE_KEY_" + identifier.upper()))
    data = {
        'name': newname,
        'domain': nostdressdomain,
        'kind': "lnbits",
        'host': os.getenv("LNBITS_HOST"),
        'key': os.getenv("LNBITS_INVOICE_KEY_" + identifier.upper()),
        'pin': pin,
        'npub': npub,
        'currentname': currentname
    }
    try:
        url = "https://" + nostdressdomain + "/api/easy/"
        res = requests.post(url, data=data)
        print(res.text)
        obj = json.loads(res.text)

        if obj.get("ok"):
            return data["name"] + "@" + nostdressdomain, obj["pin"]

    except Exception as e:
        print("Creating random name..")
        data["name"] = data["name"] + "_" + randomword(10)
        try:
            url = "https://" + nostdressdomain + "/api/easy/"
            res = requests.post(url, data=data)
            print(res.text)
            obj = json.loads(res.text)

            if obj.get("ok"):
                return data["name"] + "@" + nostdressdomain, obj["pin"]

        except Exception as e:
            return "", ""


def check_and_set_ln_bits_keys(identifier, npub):
    if not os.getenv("LNBITS_INVOICE_KEY_" + identifier.upper()):
        invoicekey, adminkey, walletid, userid, success = create_lnbits_account(identifier)
        add_key_to_env_file("LNBITS_INVOICE_KEY_" + identifier.upper(), invoicekey)
        add_key_to_env_file("LNBITS_ADMIN_KEY_" + identifier.upper(), adminkey)
        add_key_to_env_file("LNBITS_WALLET_ID_" + identifier.upper(), userid)

        lnaddress = ""
        pin = ""
        if os.getenv("NOSTDRESS_DOMAIN") and success != "failed":
            print(os.getenv("NOSTDRESS_DOMAIN"))
            lnaddress, pin = make_ln_address_nostdress(identifier, npub, " ", os.getenv("NOSTDRESS_DOMAIN"), identifier)
        add_key_to_env_file("LNADDRESS_" + identifier.upper(), lnaddress)
        add_key_to_env_file("LNADDRESS_PIN_" + identifier.upper(), pin)

        return invoicekey, adminkey, walletid, lnaddress
    else:
        return (os.getenv("LNBITS_INVOICE_KEY_" + identifier.upper()),
                os.getenv("LNBITS_ADMIN_KEY_" + identifier.upper()),
                os.getenv("LNBITS_WALLET_ID_" + identifier.upper()),
                os.getenv("LNADDRESS_" + identifier.upper()))


async def change_ln_address(identifier, new_identifier, dvm_config, updateprofile=False):
    previous_identifier = os.getenv("LNADDRESS_" + identifier.upper()).split("@")[0]
    pin = os.getenv("LNADDRESS_PIN_" + identifier.upper())
    npub = Keys.parse(os.getenv("DVM_PRIVATE_KEY_" + identifier.upper())).public_key().to_hex()
    lnaddress, pin = make_ln_address_nostdress(identifier, npub, pin, os.getenv("NOSTDRESS_DOMAIN"), new_identifier,
                                               currentname=previous_identifier)
    add_key_to_env_file("LNADDRESS_" + identifier.upper(), lnaddress)
    add_key_to_env_file("LNADDRESS_PIN_" + identifier.upper(), pin)
    print("changed lnaddress")
    if updateprofile:
        private_key = os.getenv("DVM_PRIVATE_KEY_" + identifier.upper())
        await update_profile_lnaddress(private_key, dvm_config, lud16=lnaddress)


def add_key_to_env_file(value, oskey):
    env_path = Path('.env')
    if env_path.is_file():
        dotenv.load_dotenv(env_path, verbose=True, override=True)
        dotenv.set_key(env_path, value, oskey)



""" WhisperX
"""
# We follow Semantic Versioning (https://semver.org/)
_MAJOR_VERSION = '1'
_MINOR_VERSION = '0'
_PATCH_VERSION = '1'

__version__ = '.'.join([
    _MAJOR_VERSION,
    _MINOR_VERSION,
    _PATCH_VERSION,
])



import json
import os
from datetime import timedelta

from nostr_sdk import Tag, Kind, init_logger, LogLevel, Filter, Client, NostrSigner, Keys, \
    SecretKey, Options, SingleLetterTag, Alphabet, PublicKey

from nostr_dvm.interfaces.dvmtaskinterface import DVMTaskInterface, process_venv
from nostr_dvm.utils.admin_utils import AdminConfig
from nostr_dvm.utils.definitions import EventDefinitions, relay_timeout_long
from nostr_dvm.utils.dvmconfig import DVMConfig, build_default_config
from nostr_dvm.utils.nip88_utils import NIP88Config
from nostr_dvm.utils.nip89_utils import NIP89Config, check_and_set_d_tag
from nostr_dvm.utils.output_utils import post_process_list_to_events

"""
This File contains a Module to search for notes
Accepted Inputs: a search query
Outputs: A list of events 
Params:  None
"""


class TrendingNotesGleasonator(DVMTaskInterface):
    KIND: Kind = EventDefinitions.KIND_NIP90_CONTENT_DISCOVERY
    TASK: str = "trending-content"
    FIX_COST: float = 0
    dvm_config: DVMConfig
    logger = False

    async def init_dvm(self, name, dvm_config: DVMConfig, nip89config: NIP89Config, nip88config: NIP88Config = None,
                       admin_config: AdminConfig = None, options=None):
        dvm_config.SCRIPT = os.path.abspath(__file__)

        if self.options is not None:
            if self.options.get("logger"):
                self.logger = bool(self.options.get("logger"))

            if self.logger:
                init_logger(LogLevel.DEBUG)

    async def is_input_supported(self, tags, client=None, dvm_config=None):
        for tag in tags:
            if tag.as_vec()[0] == 'i':
                input_value = tag.as_vec()[1]
                input_type = tag.as_vec()[2]
                if input_type != "text":
                    return False
        return True

    async def create_request_from_nostr_event(self, event, client=None, dvm_config=None):
        request_form = {"jobID": event.id().to_hex()}
        max_results = 200

        for tag in event.tags():
            if tag.as_vec()[0] == 'i':
                input_type = tag.as_vec()[2]
            elif tag.as_vec()[0] == 'param':
                param = tag.as_vec()[1]
                if param == "max_results":  # check for param type
                    max_results = int(tag.as_vec()[2])

        options = {
            "max_results": max_results,
            "relay": "wss://gleasonator.de/relay"
        }
        request_form['options'] = json.dumps(options)
        return request_form

    async def process(self, request_form):
        options = self.set_options(request_form)

        opts = (Options().wait_for_send(False).send_timeout(timedelta(seconds=self.dvm_config.RELAY_TIMEOUT)))
        sk = SecretKey.from_hex(self.dvm_config.PRIVATE_KEY)
        keys = Keys.parse(sk.to_hex())
        signer = NostrSigner.keys(keys)
        cli = Client.with_opts(signer, opts)

        await cli.add_relay(options["relay"])
        await cli.connect()

        ltags = ["#e", "pub.ditto.trends"]
        authors = [PublicKey.parse("db0e60d10b9555a39050c258d460c5c461f6d18f467aa9f62de1a728b8a891a4")]
        notes_filter = Filter().authors(authors).kind(Kind(1985)).custom_tag(SingleLetterTag.lowercase(Alphabet.L),
                                                                             ltags)

        events = await cli.get_events_of([notes_filter], relay_timeout_long)

        result_list = []
        if len(events) > 0:
            event = events[0]
            print(event)
            for tag in event.tags():
                if tag.as_vec()[0] == "e":
                    e_tag = Tag.parse(["e", tag.as_vec()[1], tag.as_vec()[2]])
                    result_list.append(e_tag.as_vec())

        else:
            print("Nothing found")
            return ""

        await cli.shutdown()
        print(json.dumps(result_list))
        return json.dumps(result_list)

    async def post_process(self, result, event):
        """Overwrite the interface function to return a social client readable format, if requested"""
        for tag in event.tags():
            if tag.as_vec()[0] == 'output':
                format = tag.as_vec()[1]
                if format == "text/plain":  # check for output type
                    result = post_process_list_to_events(result)

        # if not text/plain, don't post-process
        return result


# We build an example here that we can call by either calling this file directly from the main directory,
# or by adding it to our playground. You can call the example and adjust it to your needs or redefine it in the
# playground or elsewhere
def build_example(name, identifier, admin_config, custom_processing_msg):
    dvm_config = build_default_config(identifier)
    dvm_config.USE_OWN_VENV = False
    dvm_config.CUSTOM_PROCESSING_MESSAGE = custom_processing_msg
    admin_config.LUD16 = dvm_config.LN_ADDRESS
    # Add NIP89

    nip89info = {
        "name": name,
        "picture": "0c760b3ecdbc993ba47b785d0adecf00c760b3ecdbc993ba47b785d0adecf0ec71fd9c59808e27d0665b9f77a32d8de.png",
        "image": "0c760b3ecdbc993ba47b785d0adecf00c760b3ecdbc993ba47b785d0adecf0ec71fd9c59808e27d0665b9f77a32d8de.png",
        "about": "I show trending notes from Soapbox Ditto",
        "amount": "Free",
        "encryptionSupported": True,
        "cashuAccepted": True,
        "nip90Params": {}
    }
    nip89config = NIP89Config()
    nip89config.DTAG = check_and_set_d_tag(identifier, name, dvm_config.PRIVATE_KEY, nip89info["image"])
    nip89config.CONTENT = json.dumps(nip89info)

    return TrendingNotesGleasonator(name=name, dvm_config=dvm_config, nip89config=nip89config,
                                    admin_config=admin_config)


if __name__ == '__main__':
    process_venv(TrendingNotesGleasonator)



"""WhisperX Module
"""
import sys

from nova_utils.interfaces.server_module import Processor

# Setting defaults
_default_options = {"model": "tiny", "alignment_mode": "segment", "batch_size": "16", 'language': None,
                    'compute_type': 'float16'}


# supported language codes, cf. whisperx/alignment.py
# DEFAULT_ALIGN_MODELS_TORCH.keys() | DEFAULT_ALIGN_MODELS_HF.keys() | {None}
# {'vi', 'uk', 'pl', 'ur', 'ru', 'ko', 'en', 'zh', 'es', 'it', 'el', 'te', 'da', 'he', 'fa', 'pt', 'de',
#  'fr', 'tr', 'nl', 'cs', 'hu', 'fi', 'ar', 'ja', None}

class WhisperX(Processor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.options = _default_options | self.options
        self.device = None
        self.ds_iter = None
        self.session_manager = None

        # IO shortcuts
        self.input = [x for x in self.model_io if x.io_type == "input"]
        self.output = [x for x in self.model_io if x.io_type == "output"]
        assert len(self.input) == 1 and len(self.output) == 1
        self.input = self.input[0]
        self.output = self.output[0]

    def process_data(self, ds_manager) -> dict:
        import whisperx
        import torch
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        self.session_manager = self.get_session_manager(ds_manager)
        input_audio = self.session_manager.input_data['audio']

        # sliding window will be applied by WhisperX
        audio = whisperx.load_audio(input_audio.meta_data.file_path)

        # transcribe with original whisper
        try:
            model = whisperx.load_model(self.options["model"], self.device, compute_type=self.options['compute_type'],
                                        language=self.options['language'])
        except ValueError:
            print(f'Your hardware does not support {self.options["compute_type"]} - fallback to float32')
            sys.stdout.flush()
            model = whisperx.load_model(self.options["model"], self.device, compute_type='float32',
                                        language=self.options['language'])

        result = model.transcribe(audio, batch_size=int(self.options["batch_size"]))

        # delete model if low on GPU resources
        import gc;
        gc.collect();
        torch.cuda.empty_cache();
        del model

        if not self.options["alignment_mode"] == "raw":
            # load alignment model and metadata
            model_a, metadata = whisperx.load_align_model(
                language_code=result["language"], device=self.device
            )

            # align whisper output
            result_aligned = whisperx.align(
                result["segments"], model_a, metadata, audio, self.device
            )
            result = result_aligned

            # delete model if low on GPU resources
            import gc;
            gc.collect();
            torch.cuda.empty_cache();
            del model_a

        return result

    def to_output(self, data: dict):
        def _fix_missing_timestamps(data):
            """
            https://github.com/m-bain/whisperX/issues/253
            Some characters might miss timestamps and recognition scores. This function adds estimated time stamps assuming a fixed time per character of 65ms.
            Confidence for each added timestamp will be 0.
            Args:
                data (dictionary): output dictionary as returned by process_data
            """
            last_end = 0
            for s in data["segments"]:
                for w in s["words"]:
                    if "end" in w.keys():
                        last_end = w["end"]
                    else:
                        # TODO: rethink lower bound for confidence; place word centred instead of left aligned
                        w["start"] = last_end
                        last_end += 0.065
                        w["end"] = last_end
                        # w["score"] = 0.000
                        w['score'] = _hmean([x['score'] for x in s['words'] if len(x) == 4])

        def _hmean(scores):
            if len(scores) > 0:
                prod = scores[0]
                for s in scores[1:]:
                    prod *= s
                prod = prod ** (1 / len(scores))
            else:
                prod = 0
            return prod

        if (
                self.options["alignment_mode"] == "word"
                or self.options["alignment_mode"] == "segment"
        ):
            _fix_missing_timestamps(data)

        if self.options["alignment_mode"] == "word":
            anno_data = [
                (w["start"], w["end"], w["word"], w["score"])
                for w in data["word_segments"]
            ]
        else:
            anno_data = [
                # (w["start"], w["end"], w["text"], _hmean([x['score'] for x in w['words']])) for w in data["segments"]
                (w["start"], w["end"], w["text"], 1) for w in data["segments"]
                # alignment 'raw' no longer contains a score(?)
            ]

        # convert to milliseconds
        anno_data = [(x[0] * 1000, x[1] * 1000, x[2], x[3]) for x in anno_data]
        out = self.session_manager.output_data_templates[self.output.io_id]
        out.data = anno_data
        return self.session_manager.output_data_templates



import asyncio
import json
import math
import os
import signal
from datetime import timedelta

from nostr_sdk import (Keys, Client, Timestamp, Filter, nip04_decrypt, HandleNotification, EventBuilder, PublicKey,
                       Options, Tag, Event, nip04_encrypt, NostrSigner, EventId)

from nostr_dvm.utils.database_utils import fetch_user_metadata
from nostr_dvm.utils.definitions import EventDefinitions, relay_timeout
from nostr_dvm.utils.dvmconfig import DVMConfig
from nostr_dvm.utils.nip88_utils import nip88_has_active_subscription
from nostr_dvm.utils.nip89_utils import NIP89Config
from nostr_dvm.utils.nostr_utils import send_event
from nostr_dvm.utils.nwc_tools import nwc_zap
from nostr_dvm.utils.subscription_utils import create_subscription_sql_table, add_to_subscription_sql_table, \
    get_from_subscription_sql_table, update_subscription_sql_table, get_all_subscriptions_from_sql_table, \
    delete_from_subscription_sql_table
from nostr_dvm.utils.zap_utils import zaprequest


class Subscription:
    job_list: list

    # This is a simple list just to keep track which events we created and manage, so we don't pay for other requests
    def __init__(self, dvm_config, admin_config=None):
        asyncio.run(self.run_subscription(dvm_config, admin_config))

    async def run_subscription(self, dvm_config, admin_config):

        self.NAME = "Subscription Handler"
        dvm_config.DB = "db/" + "subscriptions" + ".db"
        self.dvm_config = dvm_config
        nip89config = NIP89Config()
        nip89config.NAME = self.NAME
        self.dvm_config.NIP89 = nip89config
        self.admin_config = admin_config
        self.keys = Keys.parse(dvm_config.PRIVATE_KEY)

        wait_for_send = False
        skip_disconnected_relays = True
        opts = (Options().wait_for_send(wait_for_send).send_timeout(timedelta(seconds=self.dvm_config.RELAY_TIMEOUT))
                .skip_disconnected_relays(skip_disconnected_relays))
        signer = NostrSigner.keys(self.keys)
        self.client = Client.with_opts(signer, opts)

        pk = self.keys.public_key()

        self.job_list = []

        print("Nostr Subscription Handler public key: " + str(pk.to_bech32()) + " Hex: " + str(
            pk.to_hex()) + "\n")

        for relay in self.dvm_config.RELAY_LIST:
            await self.client.add_relay(relay)
        await self.client.connect()

        zap_filter = Filter().pubkey(pk).kinds([EventDefinitions.KIND_ZAP]).since(Timestamp.now())
        cancel_subscription_filter = Filter().kinds([EventDefinitions.KIND_NIP88_STOP_SUBSCRIPTION_EVENT]).since(
            Timestamp.now())
        authors = []
        if admin_config is not None and len(admin_config.USERNPUBS) > 0:
            # we might want to limit which services can connect to the subscription handler
            for key in admin_config.USERNPUBS:
                authors.append(PublicKey.parse(key))
            dvm_filter = Filter().authors(authors).pubkey(pk).kinds(
                [EventDefinitions.KIND_NIP90_DVM_SUBSCRIPTION]).since(
                Timestamp.now())
        else:
            # or we don't
            dvm_filter = Filter().pubkey(pk).kinds(
                [EventDefinitions.KIND_NIP90_DVM_SUBSCRIPTION]).since(
                Timestamp.now())

        await self.client.subscribe([zap_filter, dvm_filter, cancel_subscription_filter], None)

        create_subscription_sql_table(dvm_config.DB)

        class NotificationHandler(HandleNotification):
            client = self.client
            dvm_config = self.dvm_config
            keys = self.keys

            async def handle(self, relay_url, subscription_id, nostr_event: Event):
                if nostr_event.kind() == EventDefinitions.KIND_NIP90_DVM_SUBSCRIPTION:
                    await handle_nwc_request(nostr_event)
                elif nostr_event.kind() == EventDefinitions.KIND_NIP88_STOP_SUBSCRIPTION_EVENT:
                    await handle_cancel(nostr_event)

            async def handle_msg(self, relay_url, msg):
                return

        async def handle_cancel(nostr_event):
            print(nostr_event.as_json())
            sender = nostr_event.author().to_hex()
            kind7001eventid = ""
            recipient = ""
            if sender == self.keys.public_key().to_hex():
                return

            for tag in nostr_event.tags():
                if tag.as_vec()[0] == "p":
                    recipient = tag.as_vec()[1]
                elif tag.as_vec()[0] == "e":
                    kind7001eventid = tag.as_vec()[1]

            if kind7001eventid != "":
                subscription = get_from_subscription_sql_table(dvm_config.DB, kind7001eventid)

                if subscription is not None:
                    update_subscription_sql_table(dvm_config.DB, kind7001eventid, recipient,
                                                  subscription.subscriber, subscription.nwc, subscription.cadence,
                                                  subscription.amount, subscription.unit, subscription.begin,
                                                  subscription.end,
                                                  subscription.tier_dtag, subscription.zaps, subscription.recipe,
                                                  False, Timestamp.now().as_secs(), subscription.tier)

        # send_status_canceled(kind7001eventid, nostr_event) # TODO, waiting for spec

        def infer_subscription_end_time(start, cadence):
            end = start
            if cadence == "daily":
                end = start + 60 * 60 * 24
            elif cadence == "weekly":
                end = start + 60 * 60 * 24 * 7
            elif cadence == "monthly":
                # TODO check days of month -.-
                end = start + 60 * 60 * 24 * 31
            elif cadence == "yearly":
                # TODO check extra day every 4 years
                end = start + 60 * 60 * 24 * 356
            return end

        async def send_status_success(original_event, domain):

            e_tag = Tag.parse(["e", original_event.id().to_hex()])
            p_tag = Tag.parse(["p", original_event.author().to_hex()])
            status_tag = Tag.parse(["status", "success", "Job has been scheduled, you can manage it on " + domain])
            reply_tags = [status_tag]
            encryption_tags = []

            encrypted_tag = Tag.parse(["encrypted"])
            encryption_tags.append(encrypted_tag)
            encryption_tags.append(p_tag)
            encryption_tags.append(e_tag)

            str_tags = []
            for element in reply_tags:
                str_tags.append(element.as_vec())

            content = json.dumps(str_tags)
            content = nip04_encrypt(self.keys.secret_key(), PublicKey.from_hex(original_event.author().to_hex()),
                                    content)
            reply_tags = encryption_tags

            keys = Keys.parse(dvm_config.PRIVATE_KEY)
            reaction_event = EventBuilder(EventDefinitions.KIND_FEEDBACK, str(content), reply_tags).to_event(keys)
            await send_event(reaction_event, client=self.client, dvm_config=self.dvm_config)
            print("[" + self.dvm_config.NIP89.NAME + "]" + ": Sent Kind " + str(
                EventDefinitions.KIND_FEEDBACK.as_u16()) + " Reaction: " + "success" + " " + reaction_event.as_json())

        async def pay_zap_split(nwc, overall_amount, zaps, tier, unit="msats"):
            overallsplit = 0

            for zap in zaps:
                print(zap)
                overallsplit += int(zap[3])

            print(overallsplit)
            zapped_amount = 0
            for zap in zaps:
                if zap[1] == "":
                    # If the client did decide to not add itself to the zap split, or if a slot is left we add the subscription service in the empty space
                    zap[1] = Keys.parse(self.dvm_config.PRIVATE_KEY).public_key().to_hex()

                name, nip05, lud16 = await fetch_user_metadata(zap[1], self.client)
                splitted_amount = math.floor(
                    (int(zap[3]) / overallsplit) * int(overall_amount) / 1000)
                # invoice = create_bolt11_lud16(lud16, splitted_amount)
                # TODO add details about DVM in message

                invoice = zaprequest(lud16, splitted_amount, tier, None,
                                     PublicKey.parse(zap[1]), self.keys, DVMConfig.RELAY_LIST)
                print(invoice)
                if invoice is not None:
                    nwc_event_id = await nwc_zap(nwc, invoice, self.keys, zap[2])
                    if nwc_event_id is None:
                        print("error zapping " + lud16)
                    else:
                        zapped_amount = zapped_amount + (splitted_amount * 1000)
                        print(str(zapped_amount) + "/" + str(overall_amount))

            if zapped_amount < overall_amount * 0.8:  # TODO how do we handle failed zaps for some addresses? we are ok with 80% for now
                # if zapped_amount < int(zaps[0][3]):

                success = False
            else:
                success = True
                # if no active subscription exists OR the subscription ended, pay

            return success

        async def make_subscription_zap_recipe(event7001, recipient, subscriber, start, end, tier_dtag):
            message = "payed by subscription service"
            pTag = Tag.parse(["p", recipient])
            PTag = Tag.parse(["P", subscriber])
            eTag = Tag.parse(["e", event7001])
            validTag = Tag.parse(["valid", str(start), str(end)])
            tierTag = Tag.parse(["tier", tier_dtag])
            alttag = Tag.parse(["alt", "This is a NIP90 DVM Subscription Payment Recipe"])

            tags = [pTag, PTag, eTag, validTag, tierTag, alttag]

            event = EventBuilder(EventDefinitions.KIND_NIP88_PAYMENT_RECIPE,
                                 message, tags).to_event(self.keys)

            dvmconfig = DVMConfig()
            signer = NostrSigner.keys(self.keys)
            client = Client(signer)
            for relay in dvmconfig.RELAY_LIST:
                await client.add_relay(relay)
            await client.connect()
            recipe = await client.send_event(event)
            recipe_id = recipe.id
            await client.shutdown()
            recipe = recipe_id.to_hex()
            return recipe

        async def handle_nwc_request(nostr_event):
            print(nostr_event.as_json())
            sender = nostr_event.author().to_hex()
            if sender == self.keys.public_key().to_hex():
                return

            try:
                decrypted_text = nip04_decrypt(self.keys.secret_key(), nostr_event.author(), nostr_event.content())
                subscriber = ""
                nwc = ""
                try:
                    jsonevent = json.loads(decrypted_text)
                    for entry in jsonevent:
                        if entry[1] == "nwc":
                            nwc = entry[2]
                        elif entry[1] == "p":
                            subscriber = entry[2]

                    subscriptionfilter = Filter().kind(EventDefinitions.KIND_NIP88_SUBSCRIBE_EVENT).author(
                        PublicKey.parse(subscriber)).limit(1)
                    evts = await self.client.get_events_of([subscriptionfilter], relay_timeout)
                    if len(evts) > 0:
                        event7001id = evts[0].id().to_hex()
                        print(evts[0].as_json())
                        tier_dtag = ""
                        recipient = ""
                        cadence = ""
                        unit = "msats"
                        zaps = []
                        tier = "DVM"
                        overall_amount = 0
                        subscription_event_id = ""
                        for tag in evts[0].tags():
                            if tag.as_vec()[0] == "amount":
                                overall_amount = int(tag.as_vec()[1])

                                unit = tag.as_vec()[2]
                                cadence = tag.as_vec()[3]
                                print(str(overall_amount) + " " + unit + " " + cadence)
                            elif tag.as_vec()[0] == "p":
                                recipient = tag.as_vec()[1]
                            elif tag.as_vec()[0] == "e":
                                subscription_event_id = tag.as_vec()[1]
                            elif tag.as_vec()[0] == "event":
                                jsonevent = json.loads(tag.as_vec()[1])
                                subscription_event = Event.from_json(jsonevent)

                                for tag in subscription_event.tags():
                                    if tag.as_vec()[0] == "d":
                                        tier_dtag = tag.as_vec()[1]
                                    elif tag.as_vec()[0] == "zap":
                                        zaps.append(tag.as_vec())
                                    elif tag.as_vec()[0] == "title":
                                        tier = tag.as_vec()[1]

                        if tier_dtag == "" or len(zaps) == 0:
                            tierfilter = Filter().id(EventId.parse(subscription_event_id))
                            evts = await self.client.get_events_of([tierfilter], relay_timeout)
                            if len(evts) > 0:
                                for tag in evts[0].tags():
                                    if tag.as_vec()[0] == "d":
                                        tier_dtag = tag.as_vec()[0]

                        isactivesubscription = False
                        recipe = ""
                        subscription = get_from_subscription_sql_table(dvm_config.DB, event7001id)

                        zapsstr = json.dumps(zaps)
                        print(zapsstr)
                        success = True
                        if subscription is None or subscription.end <= Timestamp.now().as_secs():
                            # rather check nostr if our db is right
                            subscription_status = await nip88_has_active_subscription(
                                PublicKey.parse(subscriber),
                                tier_dtag, self.client, recipient, checkCanceled=False)

                            if not subscription_status["isActive"]:
                                start = Timestamp.now().as_secs()
                                end = infer_subscription_end_time(start, cadence)

                                # we add or update the subscription in the db, with non-active subscription to avoid double payments
                                if subscription is None:
                                    add_to_subscription_sql_table(dvm_config.DB, event7001id, recipient, subscriber,
                                                                  nwc,
                                                                  cadence, overall_amount, unit, start, end, tier_dtag,
                                                                  zapsstr, recipe, isactivesubscription,
                                                                  Timestamp.now().as_secs(), tier)
                                    print("new subscription entry before payment")
                                else:
                                    update_subscription_sql_table(dvm_config.DB, event7001id, recipient, subscriber,
                                                                  nwc,
                                                                  cadence, overall_amount, unit, start, end,
                                                                  tier_dtag, zapsstr, recipe, isactivesubscription,
                                                                  Timestamp.now().as_secs(), tier)
                                    print("updated subscription entry before payment")

                                # we attempt to pay the subscription
                                success = await pay_zap_split(nwc, overall_amount, zaps, tier, unit)

                            else:
                                start = Timestamp.now().as_secs()
                                end = subscription_status["validUntil"]
                        else:
                            start = subscription.begin
                            end = subscription.end

                        if success:
                            # we create a payment recipe
                            recipe = await make_subscription_zap_recipe(event7001id, recipient, subscriber, start, end,
                                                                        tier_dtag)
                            print("RECIPE " + recipe)
                            isactivesubscription = True

                            # we then update the subscription based on payment success
                            update_subscription_sql_table(dvm_config.DB, event7001id, recipient, subscriber, nwc,
                                                          cadence, overall_amount, unit, start, end,
                                                          tier_dtag, zapsstr, recipe, isactivesubscription,
                                                          Timestamp.now().as_secs(), tier)
                            print("updated subscription entry after payment")

                            await send_status_success(nostr_event, "noogle.lol")

                            message = ("Subscribed to DVM " + tier + ". Renewing on: " + str(
                                Timestamp.from_secs(end).to_human_datetime().replace("Z", " ").replace("T",
                                                                                                       " ") + " GMT"))

                            self.client.send_private_msg(PublicKey.parse(subscriber), message, None)




                except Exception as e:
                    print(e)

            except Exception as e:
                print("Error in Subscriber " + str(e))

        def handle_expired_subscription(subscription):
            delete_threshold = 60 * 60 * 24 * 365
            if subscription.cadence == "daily":
                delete_threshold = 60 * 60 * 24 * 3  # After 3 days, delete the subscription, user can make a new one
            elif subscription.cadence == "weekly":
                delete_threshold = 60 * 60 * 24 * 21  # After 21 days, delete the subscription, user can make a new one
            elif subscription.cadence == "monthly":
                delete_threshold = 60 * 60 * 24 * 60  # After 60 days, delete the subscription, user can make a new one
            elif subscription.cadence == "yearly":
                delete_threshold = 60 * 60 * 24 * 500  # After 500 days, delete the subscription, user can make a new one

            if subscription.end < (Timestamp.now().as_secs() - delete_threshold):
                delete_from_subscription_sql_table(dvm_config.DB, subscription.id)
                print("Delete expired subscription")

        async def handle_subscription_renewal(subscription):
            zaps = json.loads(subscription.zaps)

            success = await pay_zap_split(subscription.nwc, subscription.amount, zaps, subscription.tier,
                                          subscription.unit)
            if success:
                end = infer_subscription_end_time(Timestamp.now().as_secs(), subscription.cadence)
                recipe = await make_subscription_zap_recipe(subscription.id, subscription.recipent,
                                                            subscription.subscriber, subscription.begin,
                                                            end, subscription.tier_dtag)
            else:
                end = Timestamp.now().as_secs()
                recipe = subscription.recipe
                delete_from_subscription_sql_table(dvm_config.DB, subscription.id)

            update_subscription_sql_table(dvm_config.DB, subscription.id,
                                          subscription.recipent,
                                          subscription.subscriber, subscription.nwc,
                                          subscription.cadence, subscription.amount,
                                          subscription.unit,
                                          subscription.begin, end,
                                          subscription.tier_dtag, subscription.zaps, recipe,
                                          success,
                                          Timestamp.now().as_secs(), subscription.tier)

            print("updated subscription entry")

            keys = Keys.parse(dvm_config.PRIVATE_KEY)
            message = (
                    "Renewed Subscription to DVM " + subscription.tier + ". Next renewal: " + str(
                Timestamp.from_secs(end).to_human_datetime().replace("Z", " ").replace("T",
                                                                                       " ")))
            # await self.client.send_direct_msg(PublicKey.parse(subscription.subscriber), message, None)
            await self.client.send_private_msg(PublicKey.parse(subscription.subscriber), message, None)

        async def check_subscriptions():
            try:
                subscriptions = get_all_subscriptions_from_sql_table(dvm_config.DB)

                for subscription in subscriptions:

                    if subscription.nwc == "":
                        delete_from_subscription_sql_table(dvm_config.DB, subscription.id)

                    if subscription.active:
                        if subscription.end < Timestamp.now().as_secs():
                            # We could directly zap, but let's make another check if our subscription expired
                            subscription_status = await nip88_has_active_subscription(
                                PublicKey.parse(subscription.subscriber),
                                subscription.tier_dtag, self.client, subscription.recipent)

                            if subscription_status["expires"]:
                                # if subscription expires, just note it as not active
                                update_subscription_sql_table(dvm_config.DB, subscription_status["subscriptionId"],
                                                              subscription.recipent,
                                                              subscription.subscriber, subscription.nwc,
                                                              subscription.cadence, subscription.amount,
                                                              subscription.unit,
                                                              subscription.begin, subscription.end,
                                                              subscription.tier_dtag, subscription.zaps,
                                                              subscription.recipe,
                                                              False,
                                                              Timestamp.now().as_secs(), subscription.tier)
                            else:
                                await handle_subscription_renewal(subscription)

                    else:
                        handle_expired_subscription(subscription)

                print(str(Timestamp.now().as_secs()) + ": Checking " + str(
                    len(subscriptions)) + " Subscription entries..")
            except Exception as e:
                print(e)

        asyncio.create_task(self.client.handle_notifications(NotificationHandler()))

        try:
            while True:
                await asyncio.sleep(60.0)
                await check_subscriptions()
        except KeyboardInterrupt:
            print('Stay weird!')
            os.kill(os.getpid(), signal.SIGTERM)



import os
import signal
import time

import requests
from nostr_sdk import Event

from nostr_dvm.utils.definitions import EventDefinitions
from nostr_dvm.utils.mediasource_utils import check_source_type, media_source
from nostr_dvm.utils.nostr_utils import get_event_by_id, get_referenced_event_by_id


async def get_task(event, client, dvm_config):
    try:
        if event.kind() == EventDefinitions.KIND_NIP90_GENERIC:  # use this for events that have no id yet, inclufr j tag
            for tag in event.tags():
                if tag.as_vec()[0] == 'j':
                    return tag.as_vec()[1]
            else:
                return "unknown job: " + event.as_json()
        elif event.kind() == EventDefinitions.KIND_DM:  # dm
            for tag in event.tags():
                if tag.as_vec()[0] == 'j':
                    return tag.as_vec()[1]
            else:
                return "unknown job: " + event.as_json()

        # This looks a bit more complicated, but we do several tasks for text-extraction in the future
        elif event.kind() == EventDefinitions.KIND_NIP90_EXTRACT_TEXT:
            for tag in event.tags():
                if tag.as_vec()[0] == "i":
                    if tag.as_vec()[2] == "url":
                        file_type = check_url_is_readable(tag.as_vec()[1])
                        print(file_type)
                        if file_type == "pdf":
                            return "pdf-to-text"
                        elif file_type == "audio" or file_type == "video":
                            return "speech-to-text"
                        elif file_type == "image":
                            return "image-to-text"
                        else:
                            return "unknown job"
                    elif tag.as_vec()[2] == "event":
                        evt = await get_event_by_id(tag.as_vec()[1], client=client, config=dvm_config)
                        if evt is not None:
                            if evt.kind() == 1063:
                                for tg in evt.tags():
                                    if tg.as_vec()[0] == 'url':
                                        file_type = check_url_is_readable(tg.as_vec()[1])
                                        if file_type == "pdf":
                                            return "pdf-to-text"
                                        elif file_type == "audio" or file_type == "video":
                                            return "speech-to-text"
                                        else:
                                            return "unknown job"
                            else:
                                return "unknown type"
                    else:
                        return "unknown job"
        elif event.kind() == EventDefinitions.KIND_NIP90_GENERATE_IMAGE:
            has_image_tag = False
            has_text_tag = False
            for tag in event.tags():
                if tag.as_vec()[0] == "i":
                    if tag.as_vec()[2] == "url":
                        file_type = check_url_is_readable(tag.as_vec()[1])
                        if file_type == "image":
                            has_image_tag = True
                            print("found image tag")
                    elif tag.as_vec()[2] == "job":
                        evt = await get_referenced_event_by_id(event_id=tag.as_vec()[1], kinds=
                        [EventDefinitions.KIND_NIP90_RESULT_EXTRACT_TEXT,
                         EventDefinitions.KIND_NIP90_RESULT_TRANSLATE_TEXT,
                         EventDefinitions.KIND_NIP90_RESULT_SUMMARIZE_TEXT],
                                                               client=client,
                                                               dvm_config=dvm_config)
                        if evt is not None:
                            file_type = check_url_is_readable(evt.content())
                            if file_type == "image":
                                has_image_tag = True
                    elif tag.as_vec()[2] == "text":
                        has_text_tag = True

            if has_image_tag:
                return "image-to-image"
            elif has_text_tag and not has_image_tag:
                return "text-to-image"
        #  TODO if a task can consist of multiple inputs add them here
        #  This is not ideal. Maybe such events should have their own kind

        #  else if kind is supported, simply return task
        else:

            for dvm in dvm_config.SUPPORTED_DVMS:
                if dvm.KIND.as_u16() == event.kind().as_u16():
                    return dvm.TASK
    except Exception as e:
        print("Get task: " + str(e))

    return "unknown type"


def is_input_supported_generic(tags, client, dvm_config) -> bool:
    # Handle malformed tags, missing events etc here.
    try:
        for tag in tags:
            if tag.as_vec()[0] == 'i':
                if len(tag.as_vec()) < 3:
                    print("Job Event missing/malformed i tag, skipping..")
                    return False
                else:
                    input_value = tag.as_vec()[1]
                    input_type = tag.as_vec()[2]
                    # if input_type == "event":
                    #    evt = get_event_by_id(input_value, client=client, config=dvm_config)
                    #    if evt is None:
                    #        print("Event not found")
                    #        return False
                    # TODO check_url_is_readable might be more relevant per task in the future
                    # if input_type == 'url' and check_url_is_readable(input_value) is None:
                    #    print("Url not readable / supported")
                    #    return False

        return True
    except Exception as e:
        print("Generic input check: " + str(e))


async def check_task_is_supported(event: Event, client, config=None):
    try:
        dvm_config = config
        # Check for generic issues, event maformed, referenced event not found etc..
        if not is_input_supported_generic(event.tags(), client, dvm_config):
            return False, ""

        # See if current dvm supports the task
        task = await get_task(event, client=client, dvm_config=dvm_config)
        # if task not in (x.TASK for x in dvm_config.SUPPORTED_DVMS) and not task == "generic":
        #     return False, task
        # See if current dvm can handle input for given task
        for dvm in dvm_config.SUPPORTED_DVMS:
            if dvm.TASK == task:
                if not await dvm.is_input_supported(event.tags(), client, config):
                    return False, task
        return True, task

    except Exception as e:
        print("Check task: " + str(e))


def check_url_is_readable(url):
    if not str(url).startswith("http"):
        return None

    source = check_source_type(url)
    type = media_source(source)

    if type == "url":
        # If link is comaptible with one of these file formats, move on.
        req = requests.get(url)
        content_type = req.headers['content-type']
        if content_type == 'audio/x-wav' or str(url).endswith(".wav") or content_type == 'audio/mpeg' or str(
                url).endswith(
            ".mp3") or content_type == 'audio/ogg' or str(url).endswith(".ogg"):
            return "audio"
        elif (content_type == 'image/png' or str(url).endswith(".png") or content_type == 'image/jpg' or str(
                url).endswith(
            ".jpg") or content_type == 'image/jpeg' or str(url).endswith(".jpeg") or content_type == 'image/png' or
              str(url).endswith(".png")):
            return "image"
        elif content_type == 'video/mp4' or str(url).endswith(".mp4") or content_type == 'video/avi' or str(
                url).endswith(
            ".avi") or content_type == 'video/mov' or str(url).endswith(".mov"):
            return "video"
        elif (str(url)).endswith(".pdf"):
            return "pdf"
    else:
        return type

    # Otherwise we will not offer to do the job.
    return None


def get_amount_per_task(task, dvm_config, duration=1):
    #  duration is either static 1 (for images etc) or in seconds by default (e.g. audio/video)
    for dvm in dvm_config.SUPPORTED_DVMS:  # this is currently just one
        if dvm.TASK == task or dvm.TASK == "generic":
            amount = dvm.FIX_COST + (dvm.PER_UNIT_COST * duration)
            return amount
    else:
        print("[" + dvm_config.SUPPORTED_DVMS[
            0].NAME + "] Task " + task + " is currently not supported by this instance, skipping")
        return None


def keep_alive():
    try:
        while True:
            time.sleep(10)
    except KeyboardInterrupt:
        os.kill(os.getpid(), signal.SIGKILL)
        exit(1)



# Copyright © 2023 Apple Inc.

from dataclasses import dataclass
from typing import Tuple


@dataclass
class AutoencoderConfig:
    in_channels: int = 3
    out_channels: int = 3
    latent_channels_out: int = 8
    latent_channels_in: int = 4
    block_out_channels: Tuple[int] = (128, 256, 512, 512)
    layers_per_block: int = 2
    norm_num_groups: int = 32
    scaling_factor: float = 0.18215


@dataclass
class CLIPTextModelConfig:
    num_layers: int = 23
    model_dims: int = 1024
    num_heads: int = 16
    max_length: int = 77
    vocab_size: int = 49408


@dataclass
class UNetConfig:
    in_channels: int = 4
    out_channels: int = 4
    conv_in_kernel: int = 3
    conv_out_kernel: int = 3
    block_out_channels: Tuple[int] = (320, 640, 1280, 1280)
    layers_per_block: Tuple[int] = (2, 2, 2, 2)
    mid_block_layers: int = 2
    transformer_layers_per_block: Tuple[int] = (1, 1, 1, 1)
    num_attention_heads: Tuple[int] = (5, 10, 20, 20)
    cross_attention_dim: Tuple[int] = (1024,) * 4
    norm_num_groups: int = 32


@dataclass
class DiffusionConfig:
    beta_schedule: str = "scaled_linear"
    beta_start: float = 0.00085
    beta_end: float = 0.012
    num_train_steps: int = 1000



import json
import os
import urllib.request
from io import BytesIO

import requests
from PIL import Image
from nostr_sdk import Kind

from nostr_dvm.interfaces.dvmtaskinterface import DVMTaskInterface, process_venv
from nostr_dvm.utils.admin_utils import AdminConfig
from nostr_dvm.utils.definitions import EventDefinitions
from nostr_dvm.utils.dvmconfig import DVMConfig, build_default_config
from nostr_dvm.utils.nip88_utils import NIP88Config
from nostr_dvm.utils.nip89_utils import NIP89Config, check_and_set_d_tag
from nostr_dvm.utils.output_utils import upload_media_to_hoster
from nostr_dvm.utils.zap_utils import get_price_per_sat

"""
This File contains a Module to transform an image to a short video clip using Stable Video Diffusion with replicate

Accepted Inputs: Prompt (text)
Outputs: An url to an Image
Params: 
"""


class VideoGenerationReplicateSVD(DVMTaskInterface):
    KIND: Kind = EventDefinitions.KIND_NIP90_GENERATE_VIDEO
    TASK: str = "image-to-video"
    FIX_COST: float = 120
    dependencies = [("nostr-dvm", "nostr-dvm"),
                    ("replicate", "replicate")]

    async def init_dvm(self, name, dvm_config: DVMConfig, nip89config: NIP89Config, nip88config: NIP88Config = None,
                       admin_config: AdminConfig = None, options=None):
        dvm_config.SCRIPT = os.path.abspath(__file__)

    async def is_input_supported(self, tags, client=None, dvm_config=None):
        for tag in tags:
            if tag.as_vec()[0] == 'i':
                input_value = tag.as_vec()[1]
                input_type = tag.as_vec()[2]
                if input_type != "url":
                    return False
        return True

    async def create_request_from_nostr_event(self, event, client=None, dvm_config=None):
        request_form = {"jobID": event.id().to_hex() + "_" + self.NAME.replace(" ", "")}
        url = ""
        frames = 14  # 25
        if frames == 25:
            length = "25_frames_with_svd_xt"
        else:
            length = "14_frames_with_svd"
        sizing_strategy = "maintain_aspect_ratio"  # crop_to_16_9, use_image_dimensions
        frames_per_second = 6
        motion_bucket_id = 127  # Increase overall motion in the generated video
        cond_aug = 0.02  # Amount of noise to add to input image

        for tag in event.tags():
            if tag.as_vec()[0] == 'i':
                input_type = tag.as_vec()[2]
                if input_type == "url":
                    url = str(tag.as_vec()[1]).split('#')[0]
        # TODO add params as defined above

        options = {
            "url": url,
            "length": length,
            "sizing_strategy": sizing_strategy,
            "frames_per_second": frames_per_second,
            "motion_bucket_id": motion_bucket_id,
            "cond_aug": cond_aug

        }
        request_form['options'] = json.dumps(options)

        return request_form

    async def process(self, request_form):
        try:
            options = self.set_options(request_form)
            print(options["url"])
            response = requests.get(options["url"])
            image = Image.open(BytesIO(response.content)).convert("RGB")
            image.save("./outputs/input.jpg")

            import replicate
            output = replicate.run(
                "stability-ai/stable-video-diffusion:3f0457e4619daac51203dedb472816fd4af51f3149fa7a9e0b5ffcf1b8172438",
                input={"input_image": open("./outputs/input.jpg", "rb"),
                       "video_length": options["length"],
                       "sizing_strategy": options["sizing_strategy"],
                       "frames_per_second": options["frames_per_second"],
                       "motion_bucket_id": options["motion_bucket_id"],
                       "cond_aug": options["cond_aug"]
                       }
            )
            print(output)

            urllib.request.urlretrieve(output, "./outputs/svd.mp4")
            result = await upload_media_to_hoster("./outputs/svd.mp4")
            return result

        except Exception as e:
            print("Error in Module")
            raise Exception(e)


# We build an example here that we can call by either calling this file directly from the main directory,
# or by adding it to our playground. You can call the example and adjust it to your needs or redefine it in the
# playground or elsewhere
def build_example(name, identifier, admin_config):
    dvm_config = build_default_config(identifier)
    admin_config.LUD16 = dvm_config.LN_ADDRESS
    profit_in_sats = 10
    cost_in_cent = 4.0
    dvm_config.FIX_COST = int(((cost_in_cent / (get_price_per_sat("USD") * 100)) + profit_in_sats))

    nip89info = {
        "name": name,
        "image": "https://image.nostr.build/c33ca6fc4cc038ca4adb46fdfdfda34951656f87ee364ef59095bae1495ce669.jpg",
        "about": "I use Replicate to run StableDiffusion XL",
        "encryptionSupported": True,
        "cashuAccepted": True,
        "nip90Params": {}
    }

    nip89config = NIP89Config()
    nip89config.DTAG = check_and_set_d_tag(identifier, name, dvm_config.PRIVATE_KEY, nip89info["image"])
    nip89config.CONTENT = json.dumps(nip89info)

    return VideoGenerationReplicateSVD(name=name, dvm_config=dvm_config, nip89config=nip89config,
                                       admin_config=admin_config)


if __name__ == '__main__':
    process_venv(VideoGenerationReplicateSVD)



import asyncio
import json
import os

import requests
from nostr_sdk import Kind

from nostr_dvm.interfaces.dvmtaskinterface import DVMTaskInterface, process_venv
from nostr_dvm.utils.admin_utils import AdminConfig
from nostr_dvm.utils.definitions import EventDefinitions
from nostr_dvm.utils.dvmconfig import DVMConfig, build_default_config
from nostr_dvm.utils.nip88_utils import NIP88Config
from nostr_dvm.utils.nip89_utils import NIP89Config, check_and_set_d_tag
from nostr_dvm.utils.zap_utils import get_price_per_sat

"""
This File contains a Module to transform an image to a short video clip using Stable Video Diffusion with replicate

Accepted Inputs: Prompt (text)
Outputs: An url to an Image
Params: 
"""


class AudioGenerationSonoAI(DVMTaskInterface):
    KIND: Kind = EventDefinitions.KIND_NIP90_GENERATE_MUSIC
    TASK: str = "prompt-to-music"
    FIX_COST: float = 120

    async def init_dvm(self, name, dvm_config: DVMConfig, nip89config: NIP89Config, nip88config: NIP88Config = None,
                       admin_config: AdminConfig = None, options=None):
        dvm_config.SCRIPT = os.path.abspath(__file__)
        self.base_url = 'http://localhost:3000'

    async def is_input_supported(self, tags, client=None, dvm_config=None):
        for tag in tags:
            if tag.as_vec()[0] == 'i':
                input_value = tag.as_vec()[1]
                input_type = tag.as_vec()[2]
                if input_type != "text":
                    return False
        return True

    async def create_request_from_nostr_event(self, event, client=None, dvm_config=None):
        request_form = {"jobID": event.id().to_hex() + "_" + self.NAME.replace(" ", "")}

        prompt = "A popular heavy metal song about a purple Ostrich, Nostr, sung by a deep-voiced male singer, slowly and melodiously. The lyrics depict hope for a better future."
        for tag in event.tags():
            if tag.as_vec()[0] == 'i':
                input_type = tag.as_vec()[2]
                if input_type == "text":
                    prompt = str(tag.as_vec()[1])
        # TODO add params as defined above

        options = {
            "prompt": prompt,
        }
        request_form['options'] = json.dumps(options)

        return request_form

    def custom_generate_audio(self, payload):
        url = f"{self.base_url}/api/custom_generate"
        response = requests.post(url, json=payload, headers={'Content-Type': 'application/json'})
        return response.json()

    def extend_audio(self, payload):
        url = f"{self, self.base_url}/api/extend_audio"
        response = requests.post(url, json=payload, headers={'Content-Type': 'application/json'})
        return response.json()

    def generate_audio_by_prompt(self, payload):
        url = f"{self.base_url}/api/generate"
        response = requests.post(url, json=payload, headers={'Content-Type': 'application/json'})
        return response.json()

    def get_audio_information(self, audio_ids):
        url = f"{self.base_url}/api/get?ids={audio_ids}"
        response = requests.get(url)
        return response.json()

    def get_quota_information(self):
        url = f"{self.base_url}/api/get_limit"
        response = requests.get(url)
        return response.json()

    def get_clip(self, clip_id):
        url = f"{self.base_url}/api/clip?id={clip_id}"
        response = requests.get(url)
        return response.json()

    def generate_whole_song(self, clip_id):
        payload = {"clip_id": clip_id}
        url = f"{self.base_url}/api/concat"
        response = requests.post(url, json=payload)
        return response.json()

    async def process(self, request_form):
        try:
            options = self.set_options(request_form)
            has_quota = False
            quota_info = self.get_quota_information()

            if int(quota_info['credits_left']) >= 20:
                has_quota = True
            else:
                print("No quota left, exiting.")

            if has_quota:

                data = self.generate_audio_by_prompt({
                    "prompt": options["prompt"],
                    "make_instrumental": False,
                    "wait_audio": False
                })
                if len(data) == 0:
                    print("Couldn't create song")
                    pass

                ids = f"{data[0]['id']},{data[1]['id']}"
                print(f"ids: {ids}")

                for _ in range(60):
                    data = self.get_audio_information(ids)
                    if data[0]["status"] == 'streaming':
                        print(f"{data[0]['id']} ==> {data[0]['video_url']}")
                        print(f"{data[1]['id']} ==> {data[1]['video_url']}")
                        break
                    # sleep 5s
                    await asyncio.sleep(5.0)

                response1 = self.get_clip(data[0]['id'])
                print(response1['video_url'])
                print(response1['prompt'])

                response2 = self.get_clip(data[1]['id'])
                print(response2['video_url'])
                # print(response2['prompt']) #same as 1

                return response1['video_url'] + "\n" + response2['video_url'] + "\n" + response1['prompt']

        except Exception as e:
            print("Error in Module")
            raise Exception(e)


# We build an example here that we can call by either calling this file directly from the main directory,
# or by adding it to our playground. You can call the example and adjust it to your needs or redefine it in the
# playground or elsewhere
def build_example(name, identifier, admin_config):
    dvm_config = build_default_config(identifier)
    admin_config.LUD16 = dvm_config.LN_ADDRESS
    profit_in_sats = 10
    cost_in_cent = 4.0
    dvm_config.FIX_COST = int(((cost_in_cent / (get_price_per_sat("USD") * 100)) + profit_in_sats))

    nip89info = {
        "name": name,
        "image": "https://image.nostr.build/c33ca6fc4cc038ca4adb46fdfdfda34951656f87ee364ef59095bae1495ce669.jpg",
        "about": "I create songs based on prompts with suno.ai",
        "encryptionSupported": True,
        "cashuAccepted": True,
        "nip90Params": {}
    }

    nip89config = NIP89Config()
    nip89config.DTAG = check_and_set_d_tag(identifier, name, dvm_config.PRIVATE_KEY, nip89info["image"])
    nip89config.CONTENT = json.dumps(nip89info)

    return AudioGenerationSonoAI(name=name, dvm_config=dvm_config, nip89config=nip89config,
                                 admin_config=admin_config)


if __name__ == '__main__':
    process_venv(AudioGenerationSonoAI)



import sqlite3
from dataclasses import dataclass
from sqlite3 import Error


@dataclass
class Subscription:
    id: str
    recipent: str
    subscriber: str
    nwc: str
    cadence: str
    amount: int
    unit: str
    begin: int
    end: int
    tier_dtag: str
    zaps: str
    recipe: str
    active: bool
    lastupdate: int
    tier: str


def create_subscription_sql_table(db):
    try:
        import os
        if not os.path.exists(r'db'):
            os.makedirs(r'db')
        if not os.path.exists(r'outputs'):
            os.makedirs(r'outputs')
        con = sqlite3.connect(db)
        cur = con.cursor()
        cur.execute(""" CREATE TABLE IF NOT EXISTS subscriptions (
                                            id text PRIMARY KEY,
                                            recipient text,
                                            subscriber  text,
                                            nwc text NOT NULL,
                                            cadence text,
                                            amount int,
                                            unit text,
                                            begin int,
                                            end int,
                                            tier_dtag text,
                                            zaps text,
                                            recipe text,
                                            active boolean,
                                            lastupdate int,
                                            tier text
                                            
                                          
                                        ); """)
        cur.execute("SELECT name FROM sqlite_master")
        con.close()

    except Error as e:
        print(e)


def add_to_subscription_sql_table(db, id, recipient, subscriber, nwc, cadence, amount, unit, begin, end, tier_dtag,
                                  zaps,
                                  recipe, active, lastupdate, tier):
    try:
        con = sqlite3.connect(db)
        cur = con.cursor()
        data = (
        id, recipient, subscriber, nwc, cadence, amount, unit, begin, end, tier_dtag, zaps, recipe, active, lastupdate,
        tier)
        print(id)
        print(recipient)
        print(subscriber)
        print(nwc)
        cur.execute("INSERT or IGNORE INTO subscriptions VALUES(?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)", data)
        con.commit()
        con.close()
    except Error as e:
        print("Error when Adding to DB: " + str(e))


def get_from_subscription_sql_table(db, id):
    try:
        con = sqlite3.connect(db)
        cur = con.cursor()
        cur.execute("SELECT * FROM subscriptions WHERE id=?", (id,))
        row = cur.fetchone()
        con.close()
        if row is None:
            return None
        else:
            subscription = Subscription
            subscription.id = row[0]
            subscription.recipent = row[1]
            subscription.subscriber = row[2]
            subscription.nwc = row[3]
            subscription.cadence = row[4]
            subscription.amount = row[5]
            subscription.unit = row[6]
            subscription.begin = row[7]
            subscription.end = row[8]
            subscription.tier_dtag = row[9]
            subscription.zaps = row[10]
            subscription.recipe = row[11]
            subscription.active = row[12]
            subscription.lastupdate = row[13]
            subscription.tier = row[14]

            return subscription

    except Error as e:
        print("Error Getting from DB: " + str(e))
        return None


def get_all_subscriptions_from_sql_table(db):
    try:
        con = sqlite3.connect(db)
        cursor = con.cursor()

        sqlite_select_query = """SELECT * from subscriptions"""
        cursor.execute(sqlite_select_query)
        records = cursor.fetchall()
        subscriptions = []
        for row in records:
            subscriptions.append(
                Subscription(row[0], row[1], row[2], row[3], row[4], row[5], row[6], row[7], row[8], row[9], row[10],
                             row[11], row[12], row[13], row[14]))
        cursor.close()
        return subscriptions

    except sqlite3.Error as error:
        print("Failed to read data from sqlite table", error)
    finally:
        if con:
            con.close()
            # print("The SQLite connection is closed")


def delete_from_subscription_sql_table(db, id):
    try:
        con = sqlite3.connect(db)
        cur = con.cursor()
        cur.execute("DELETE FROM subscriptions WHERE id=?", (id,))
        con.commit()
        con.close()
    except Error as e:
        print(e)


def update_subscription_sql_table(db, id, recipient, subscriber, nwc, cadence, amount, unit, begin, end, tier_dtag,
                                  zaps,
                                  recipe, active, lastupdate, tier):
    try:
        con = sqlite3.connect(db)
        cur = con.cursor()
        data = (
        recipient, subscriber, nwc, cadence, amount, unit, begin, end, tier_dtag, zaps, recipe, active, lastupdate,
        tier, id)

        cur.execute(""" UPDATE subscriptions
                  SET recipient = ? ,
                      subscriber = ? ,
                      nwc = ? ,
                      cadence = ? ,
                      amount = ? ,
                      unit = ? ,
                      begin = ? ,
                      end = ?,
                      tier_dtag = ?,
                      zaps = ?,
                      recipe = ?,
                      active = ?,
                      lastupdate = ?,
                      tier = ?

                  WHERE id = ?""", data)
        con.commit()
        con.close()
    except Error as e:
        print("Error Updating DB: " + str(e))



from nostr_sdk import Tag, Keys, EventBuilder, Kind

from nostr_dvm.utils.definitions import EventDefinitions
from nostr_dvm.utils.nostr_utils import send_event
from nostr_dvm.utils.print_utils import bcolors


async def announce_dm_relays(dvm_config, client):
    tags = []

    for relay in dvm_config.RELAY_LIST:
        r_tag = Tag.parse(["r", relay])
        tags.append(r_tag)

    keys = Keys.parse(dvm_config.NIP89.PK)
    content = ""

    event = EventBuilder(Kind(10050), content, tags).to_event(keys)
    eventid = await send_event(event, client=client, dvm_config=dvm_config)
    if (eventid is not None):
        print(
            bcolors.BLUE + "[" + dvm_config.NIP89.NAME + "] Announced DM relays for " + dvm_config.NIP89.NAME + " (EventID: " + str(
                eventid.id.to_hex()) + ")" + bcolors.ENDC)
    else:
        print(
            bcolors.RED + "[" + dvm_config.NIP89.NAME + "] Could not announce DM relays for " + dvm_config.NIP89.NAME + bcolors.ENDC)


async def nip65_announce_relays(dvm_config, client):
    # todo we might want to call the dm relays seperately but for now we do it together with the inbox relays
    await announce_dm_relays(dvm_config, client)

    tags = []

    for relay in dvm_config.RELAY_LIST:
        r_tag = Tag.parse(["r", relay])
        tags.append(r_tag)

    keys = Keys.parse(dvm_config.NIP89.PK)
    content = ""

    event = EventBuilder(EventDefinitions.KIND_RELAY_ANNOUNCEMENT, content, tags).to_event(keys)
    eventid = await send_event(event, client=client, dvm_config=dvm_config)
    if (eventid is not None):
        print(
            bcolors.BLUE + "[" + dvm_config.NIP89.NAME + "] Announced NIP 65 for " + dvm_config.NIP89.NAME + " (EventID: " + str(
                eventid.id.to_hex()) + ")" + bcolors.ENDC)
    else:
        print(
            bcolors.RED + "[" + dvm_config.NIP89.NAME + "] Could not announce NIP 65 for " + dvm_config.NIP89.NAME + bcolors.ENDC)



import json
import os
from datetime import timedelta

from nostr_sdk import Timestamp, Tag, Keys, Options, SecretKey, NostrSigner, NostrDatabase, \
    ClientBuilder, Filter, NegentropyOptions, NegentropyDirection, init_logger, LogLevel, Kind, \
    RelayLimits

from nostr_dvm.interfaces.dvmtaskinterface import DVMTaskInterface, process_venv
from nostr_dvm.utils import definitions
from nostr_dvm.utils.admin_utils import AdminConfig
from nostr_dvm.utils.definitions import EventDefinitions
from nostr_dvm.utils.dvmconfig import DVMConfig, build_default_config
from nostr_dvm.utils.nip88_utils import NIP88Config, check_and_set_d_tag_nip88, check_and_set_tiereventid_nip88
from nostr_dvm.utils.nip89_utils import NIP89Config, check_and_set_d_tag, create_amount_tag
from nostr_dvm.utils.output_utils import post_process_list_to_events

"""
This File contains a Module to discover popular notes
Accepted Inputs: none
Outputs: A list of events 
Params:  None
"""


class DicoverContentLatestLongForm(DVMTaskInterface):
    KIND: Kind = EventDefinitions.KIND_NIP90_CONTENT_DISCOVERY
    TASK: str = "discover-content"
    FIX_COST: float = 0
    dvm_config: DVMConfig
    last_schedule: int
    db_since = 3600
    db_name = "db/nostr_recent_notes.db"
    min_reactions = 2
    personalized = False
    result = ""

    async def init_dvm(self, name, dvm_config: DVMConfig, nip89config: NIP89Config, nip88config: NIP88Config = None,
                       admin_config: AdminConfig = None, options=None):
        dvm_config.SCRIPT = os.path.abspath(__file__)
        self.request_form = {"jobID": "generic"}
        opts = {
            "max_results": 200,
        }
        self.request_form['options'] = json.dumps(opts)

        self.last_schedule = Timestamp.now().as_secs()

        if self.options.get("db_name"):
            self.db_name = self.options.get("db_name")
        if self.options.get("db_since"):
            self.db_since = int(self.options.get("db_since"))

        use_logger = False
        if use_logger:
            init_logger(LogLevel.DEBUG)

        if self.dvm_config.UPDATE_DATABASE:
            await self.sync_db()

        if not self.personalized:
            self.result = await self.calculate_result(self.request_form)

    async def is_input_supported(self, tags, client=None, dvm_config=None):
        for tag in tags:
            if tag.as_vec()[0] == 'i':
                input_value = tag.as_vec()[1]
                input_type = tag.as_vec()[2]
                if input_type != "text":
                    return False
        return True

    async def create_request_from_nostr_event(self, event, client=None, dvm_config=None):
        self.dvm_config = dvm_config

        request_form = {"jobID": event.id().to_hex()}

        # default values
        search = ""
        max_results = 200

        for tag in event.tags():
            if tag.as_vec()[0] == 'i':
                input_type = tag.as_vec()[2]
            elif tag.as_vec()[0] == 'param':
                param = tag.as_vec()[1]
                if param == "max_results":  # check for param type
                    max_results = int(tag.as_vec()[2])

        options = {
            "max_results": max_results,
        }
        request_form['options'] = json.dumps(options)
        return request_form

    async def process(self, request_form):
        # if the dvm supports individual results, recalculate it every time for the request
        if self.personalized:
            return await self.calculate_result(request_form)
        # else return the result that gets updated once every scheduled update. In this case on database update.
        else:
            return self.result

    async def calculate_result(self, request_form):
        from types import SimpleNamespace
        ns = SimpleNamespace()

        options = self.set_options(request_form)

        opts = (Options().wait_for_send(False).send_timeout(timedelta(seconds=self.dvm_config.RELAY_TIMEOUT)))
        sk = SecretKey.from_hex(self.dvm_config.PRIVATE_KEY)
        keys = Keys.parse(sk.to_hex())
        signer = NostrSigner.keys(keys)

        database = NostrDatabase.lmdb(self.db_name)
        # print(self.db_name)
        cli = ClientBuilder().database(database).signer(signer).opts(opts).build()
        await cli.connect()

        # Negentropy reconciliation
        # Query events from database
        timestamp_hour_ago = Timestamp.now().as_secs() - self.db_since
        since = Timestamp.from_secs(timestamp_hour_ago)

        filter1 = Filter().kind(definitions.EventDefinitions.KIND_LONGFORM).since(since)
        events = await cli.database().query([filter1])
        if self.dvm_config.LOGLEVEL.value >= LogLevel.DEBUG.value:
            print("[" + self.dvm_config.NIP89.NAME + "] Considering " + str(len(events)) + " Events")
        ns.finallist = {}
        index = options["max_results"]
        for event in events:
            if event.created_at().as_secs() > timestamp_hour_ago:
                ns.finallist[event.id().to_hex()] = index
                index = index - 1
        if len(ns.finallist) == 0:
            await cli.shutdown()
            return self.result

        result_list = []
        finallist_sorted = sorted(ns.finallist.items(), key=lambda x: x[1], reverse=True)[:int(options["max_results"])]
        for entry in finallist_sorted:
            # print(EventId.parse(entry[0]).to_bech32() + "/" + EventId.parse(entry[0]).to_hex() + ": " + str(entry[1]))
            e_tag = Tag.parse(["e", entry[0]])
            result_list.append(e_tag.as_vec())
        await cli.shutdown()
        if self.dvm_config.LOGLEVEL.value >= LogLevel.DEBUG.value:
            print("[" + self.dvm_config.NIP89.NAME + "] Filtered " + str(
                len(result_list)) + " fitting events.")
        return json.dumps(result_list)

    async def post_process(self, result, event):
        """Overwrite the interface function to return a social client readable format, if requested"""
        for tag in event.tags():
            if tag.as_vec()[0] == 'output':
                format = tag.as_vec()[1]
                if format == "text/plain":  # check for output type
                    result = post_process_list_to_events(result)

        # if not text/plain, don't post-process
        return result

    async def schedule(self, dvm_config):
        if dvm_config.SCHEDULE_UPDATES_SECONDS == 0:
            return 0
        else:
            if Timestamp.now().as_secs() >= self.last_schedule + dvm_config.SCHEDULE_UPDATES_SECONDS:
                if self.dvm_config.UPDATE_DATABASE:
                    await self.sync_db()
                self.last_schedule = Timestamp.now().as_secs()
                self.result = await self.calculate_result(self.request_form)
                return 1

    async def sync_db(self):
        try:
            relaylimits = RelayLimits.disable()
            opts = (Options().wait_for_send(False).send_timeout(
                timedelta(seconds=self.dvm_config.RELAY_LONG_TIMEOUT))).relay_limits(relaylimits)
            sk = SecretKey.from_hex(self.dvm_config.PRIVATE_KEY)
            keys = Keys.parse(sk.to_hex())
            signer = NostrSigner.keys(keys)
            database = NostrDatabase.lmdb(self.db_name)
            cli = ClientBuilder().signer(signer).database(database).opts(opts).build()

            for relay in self.dvm_config.RECONCILE_DB_RELAY_LIST:
                await cli.add_relay(relay)

            await cli.connect()

            timestamp_since = Timestamp.now().as_secs() - self.db_since
            since = Timestamp.from_secs(timestamp_since)

            filter1 = Filter().kinds([definitions.EventDefinitions.KIND_LONGFORM]).since(
                since)  # Notes, reactions, zaps

            # filter = Filter().author(keys.public_key())
            if self.dvm_config.LOGLEVEL.value >= LogLevel.DEBUG.value:
                print("[" + self.dvm_config.NIP89.NAME + "] Syncing notes of the last " + str(
                    self.db_since) + " seconds.. this might take a while..")
            dbopts = NegentropyOptions().direction(NegentropyDirection.DOWN)
            await cli.reconcile(filter1, dbopts)
            await cli.database().delete(Filter().until(Timestamp.from_secs(
                Timestamp.now().as_secs() - self.db_since)))  # Clear old events so db doesn't get too full.
            await cli.shutdown()
            if self.dvm_config.LOGLEVEL.value >= LogLevel.DEBUG.value:
                print(
                    "[" + self.dvm_config.NIP89.NAME + "] Done Syncing Notes of the last " + str(
                        self.db_since) + " seconds..")
        except Exception as e:
            print(e)


# We build an example here that we can call by either calling this file directly from the main directory,
# or by adding it to our playground. You can call the example and adjust it to your needs or redefine it in the
# playground or elsewhere
def build_example(name, identifier, admin_config, options, cost=0, update_rate=180, processing_msg=None,
                  update_db=True):
    dvm_config = build_default_config(identifier)
    dvm_config.USE_OWN_VENV = False
    dvm_config.SHOWLOG = True
    dvm_config.SCHEDULE_UPDATES_SECONDS = update_rate  # Every 10 minutes
    dvm_config.UPDATE_DATABASE = update_db
    # Activate these to use a subscription based model instead
    # dvm_config.SUBSCRIPTION_REQUIRED = True
    # dvm_config.SUBSCRIPTION_DAILY_COST = 1
    dvm_config.FIX_COST = cost
    dvm_config.CUSTOM_PROCESSING_MESSAGE = processing_msg
    admin_config.LUD16 = dvm_config.LN_ADDRESS

    image = "https://image.nostr.build/d30a75c438a8b0815b5c65b494988da26fce719f4138058929fa52d2a2dc3433.jpg",

    # Add NIP89
    nip89info = {
        "name": name,
        "image": image,
        "picture": image,
        "about": "I show the latest longform notes.",
        "lud16": dvm_config.LN_ADDRESS,
        "encryptionSupported": True,
        "cashuAccepted": True,
        "personalized": False,
        "amount": create_amount_tag(cost),
        "nip90Params": {
            "max_results": {
                "required": False,
                "values": [],
                "description": "The number of maximum results to return (default currently 100)"
            }
        }
    }

    nip89config = NIP89Config()
    nip89config.DTAG = check_and_set_d_tag(identifier, name, dvm_config.PRIVATE_KEY, nip89info["image"])
    nip89config.CONTENT = json.dumps(nip89info)

    # admin_config.UPDATE_PROFILE = False

    # admin_config.REBROADCAST_NIP89 = False

    return DicoverContentLatestLongForm(name=name, dvm_config=dvm_config, nip89config=nip89config,
                                        admin_config=admin_config, options=options)


def build_example_subscription(name, identifier, admin_config, options, update_rate=180, processing_msg=None,
                               update_db=True):
    dvm_config = build_default_config(identifier)
    dvm_config.USE_OWN_VENV = False
    dvm_config.SHOWLOG = True
    dvm_config.SCHEDULE_UPDATES_SECONDS = update_rate  # Every 3 minutes
    dvm_config.UPDATE_DATABASE = update_db
    # Activate these to use a subscription based model instead
    # dvm_config.SUBSCRIPTION_DAILY_COST = 1
    dvm_config.FIX_COST = 0
    dvm_config.CUSTOM_PROCESSING_MESSAGE = processing_msg
    admin_config.LUD16 = dvm_config.LN_ADDRESS

    image = "https://image.nostr.build/d30a75c438a8b0815b5c65b494988da26fce719f4138058929fa52d2a2dc3433.jpg",
    # Add NIP89
    nip89info = {
        "name": name,
        "image": image,
        "picture": image,
        "about": "I show the latest longform notes",
        "lud16": dvm_config.LN_ADDRESS,
        "encryptionSupported": True,
        "cashuAccepted": True,
        "subscription": True,
        "personalized": False,
        "nip90Params": {
            "max_results": {
                "required": False,
                "values": [],
                "description": "The number of maximum results to return (default currently 100)"
            }
        }
    }

    nip89config = NIP89Config()
    nip89config.DTAG = check_and_set_d_tag(identifier, name, dvm_config.PRIVATE_KEY, nip89info["image"])
    nip89config.CONTENT = json.dumps(nip89info)

    nip88config = NIP88Config()
    nip88config.DTAG = check_and_set_d_tag_nip88(identifier, name, dvm_config.PRIVATE_KEY, nip89info["image"])
    nip88config.TIER_EVENT = check_and_set_tiereventid_nip88(identifier, "1")
    nip89config.NAME = name
    nip88config.IMAGE = nip89info["image"]
    nip88config.TITLE = name
    nip88config.AMOUNT_DAILY = 100
    nip88config.AMOUNT_MONTHLY = 2000
    nip88config.CONTENT = "Subscribe to the DVM for unlimited use during your subscription"
    nip88config.PERK1DESC = "Unlimited requests"
    nip88config.PERK2DESC = "Support NostrDVM & NostrSDK development"
    nip88config.PAYMENT_VERIFIER_PUBKEY = "5b5c045ecdf66fb540bdf2049fe0ef7f1a566fa427a4fe50d400a011b65a3a7e"

    # admin_config.UPDATE_PROFILE = False
    # admin_config.REBROADCAST_NIP89 = False
    # admin_config.REBROADCAST_NIP88 = False

    # admin_config.FETCH_NIP88 = True
    # admin_config.EVENTID = ""
    # admin_config.PRIVKEY = dvm_config.PRIVATE_KEY

    return DicoverContentLatestLongForm(name=name, dvm_config=dvm_config, nip89config=nip89config,
                                        nip88config=nip88config, options=options,
                                        admin_config=admin_config)


if __name__ == '__main__':
    process_venv(DicoverContentLatestLongForm)



"""StableDiffusionXL Module
"""

import gc
import os
import sys

# Add local dir to path for relative imports
sys.path.insert(0, os.path.dirname(__file__))

from nova_utils.interfaces.server_module import Processor
from nova_utils.utils.cache_utils import get_file
from diffusers import StableDiffusionXLImg2ImgPipeline, StableDiffusionInstructPix2PixPipeline, \
    EulerAncestralDiscreteScheduler
import numpy as np
from PIL import Image as PILImage
from lora import build_lora_xl

# Setting defaults
_default_options = {"model": "stabilityai/stable-diffusion-xl-refiner-1.0", "strength": "0.58",
                    "guidance_scale": "11.0", "n_steps": "30", "lora": "", "lora_weight": "0.5"}


# TODO: add log infos,
class StableDiffusionXL(Processor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.options = _default_options | self.options
        self.device = None
        self.ds_iter = None
        self.current_session = None

        # IO shortcuts
        self.input = [x for x in self.model_io if x.io_type == "input"]
        self.output = [x for x in self.model_io if x.io_type == "output"]
        self.input = self.input[0]
        self.output = self.output[0]

    def process_data(self, ds_iter) -> dict:
        import torch
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        self.ds_iter = ds_iter
        current_session_name = self.ds_iter.session_names[0]
        self.current_session = self.ds_iter.sessions[current_session_name]['manager']
        # input_image_url = self.current_session.input_data['input_image_url'].data
        # input_image_url = ' '.join(input_image_url)
        input_image = self.current_session.input_data['input_image'].data
        input_prompt = self.current_session.input_data['input_prompt'].data
        input_prompt = ' '.join(input_prompt)
        negative_prompt = self.current_session.input_data['negative_prompt'].data
        negative_prompt = ' '.join(negative_prompt)
        # print("Input Image: " + input_image_url)
        print("Input prompt: " + input_prompt)
        print("Negative prompt: " + negative_prompt)

        try:

            model = self.options['model']
            lora = self.options['lora']
            # init_image = load_image(input_image_url).convert("RGB")
            init_image = PILImage.fromarray(input_image)

            mwidth = 1024
            mheight = 1024
            w = mwidth
            h = mheight
            if init_image.width > init_image.height:
                scale = float(init_image.height / init_image.width)
                w = mwidth
                h = int(mheight * scale)
            elif init_image.width < init_image.height:
                scale = float(init_image.width / init_image.height)
                w = int(mwidth * scale)
                h = mheight
            else:
                w = mwidth
                h = mheight

            init_image = init_image.resize((w, h))

            if lora != "" and lora != "None":
                print("Loading lora...")

                lora, input_prompt, existing_lora = build_lora_xl(lora, input_prompt, "")

                from diffusers import AutoPipelineForImage2Image
                import torch

                # init_image = init_image.resize((int(w/2), int(h/2)))

                pipe = AutoPipelineForImage2Image.from_pretrained(
                    "stabilityai/stable-diffusion-xl-base-1.0",
                    torch_dtype=torch.float16).to("cuda")

                if existing_lora:
                    lora_uri = [x for x in self.trainer.meta_uri if x.uri_id == lora][0]
                    if str(lora_uri) == "":
                        return "Lora not found"
                    lora_path = get_file(
                        fname=str(lora_uri.uri_id) + ".safetensors",
                        origin=lora_uri.uri_url,
                        file_hash=lora_uri.uri_hash,
                        cache_dir=os.getenv("CACHE_DIR"),
                        tmp_dir=os.getenv("TMP_DIR"),
                    )
                    pipe.load_lora_weights(str(lora_path))
                    print("Loaded Lora: " + str(lora_path))

                seed = 20000
                generator = torch.manual_seed(seed)

                # os.environ["PYTORCH_CUDA_ALLOC_CONF"] = "max_split_size_mb:512"

                image = pipe(
                    prompt=input_prompt,
                    negative_prompt=negative_prompt,
                    image=init_image,
                    generator=generator,
                    num_inference_steps=int(self.options['n_steps']),
                    image_guidance_scale=float(self.options['guidance_scale']),
                    strength=float(str(self.options['strength']))).images[0]


            elif model == "stabilityai/stable-diffusion-xl-refiner-1.0":

                pipe = StableDiffusionXLImg2ImgPipeline.from_pretrained(
                    model, torch_dtype=torch.float16, variant="fp16",
                    use_safetensors=True
                )

                n_steps = int(self.options['n_steps'])
                transformation_strength = float(self.options['strength'])
                cfg_scale = float(self.options['guidance_scale'])

                pipe = pipe.to(self.device)
                image = pipe(input_prompt, image=init_image,
                             negative_prompt=negative_prompt, num_inference_steps=n_steps,
                             strength=transformation_strength, guidance_scale=cfg_scale).images[0]

            elif model == "timbrooks/instruct-pix2pix":
                pipe = StableDiffusionInstructPix2PixPipeline.from_pretrained(model, torch_dtype=torch.float16,
                                                                              safety_checker=None)

                pipe.scheduler = EulerAncestralDiscreteScheduler.from_config(pipe.scheduler.config)

                pipe.to(self.device)
                n_steps = int(self.options['n_steps'])
                cfg_scale = float(self.options['guidance_scale'])
                image = \
                pipe(input_prompt, negative_prompt=negative_prompt, image=init_image, num_inference_steps=n_steps,
                     image_guidance_scale=cfg_scale).images[0]

            if torch.cuda.is_available():
                del pipe
                gc.collect()
                torch.cuda.empty_cache()
                torch.cuda.ipc_collect()

            numpy_array = np.array(image)
            return numpy_array


        except Exception as e:
            print(e)
            sys.stdout.flush()
            return "Error"

    def to_output(self, data: dict):
        self.current_session.output_data_templates['output_image'].data = data
        return self.current_session.output_data_templates



from setuptools import setup, find_packages

VERSION = '0.9.6'
DESCRIPTION = 'A framework to build and run Nostr NIP90 Data Vending Machines'
LONG_DESCRIPTION = ('A framework to build and run Nostr NIP90 Data Vending Machines. See the github repository for more information')

# Setting up
setup(
    name="nostr-dvm",
    version=VERSION,
    author="Believethehype",
    author_email="believethehypeonnostr@proton.me",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    packages=find_packages(include=['nostr_dvm', 'nostr_dvm.*']),

    install_requires=["nostr-sdk==0.35.1",
                      "bech32==1.2.0",
                      "pycryptodome==3.20.0",
                      "yt-dlp==2024.5.27",
                      "python-dotenv==1.0.0",
                      "emoji==2.12.1",
                      "ffmpegio==0.9.1",
                      "Pillow==10.1.0",
                      "PyUpload==0.1.4",
                      "pandas==2.2.2",
                      "requests==2.32.3",
                      "moviepy==2.0.0.dev2",
                      "zipp==3.17.0",
                      "urllib3==2.2.1",
                      "networkx==3.3",
                      "scipy==1.13.1",
                      "beautifulsoup4==4.12.3"
                      ],
    keywords=['nostr', 'nip90', 'dvm', 'data vending machine'],
    url="https://github.com/believethehype/nostrdvm",
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Education",
        "Programming Language :: Python :: 3",
    ]
)


import json
import os

import ffmpegio
from nostr_sdk import Kind

from nostr_dvm.utils.nip88_utils import NIP88Config

os.environ["PYTORCH_ENABLE_MPS_FALLBACK"] = "1"
from pathlib import Path
import urllib.request

from nostr_dvm.interfaces.dvmtaskinterface import DVMTaskInterface, process_venv
from nostr_dvm.utils.admin_utils import AdminConfig
from nostr_dvm.utils.definitions import EventDefinitions
from nostr_dvm.utils.dvmconfig import DVMConfig, build_default_config
from nostr_dvm.utils.nip89_utils import NIP89Config, check_and_set_d_tag
from nostr_dvm.utils.output_utils import upload_media_to_hoster
from nostr_dvm.utils.nostr_utils import get_event_by_id, get_referenced_event_by_id

"""
This File contains a Module to generate Audio based on an input and a voice

Accepted Inputs: Text
Outputs: Generated Audiofile
"""


class TextToSpeech(DVMTaskInterface):
    KIND: Kind = EventDefinitions.KIND_NIP90_TEXT_TO_SPEECH
    TASK: str = "text-to-speech"
    FIX_COST: float = 50
    PER_UNIT_COST = 0.5
    dependencies = [("nostr-dvm", "nostr-dvm"),
                    ("TTS", "TTS==0.22.0")]

    async def init_dvm(self, name, dvm_config: DVMConfig, nip89config: NIP89Config, nip88config: NIP88Config = None,
                       admin_config: AdminConfig = None, options=None):
        dvm_config.SCRIPT = os.path.abspath(__file__)
        self.dvm_config = dvm_config

    async def is_input_supported(self, tags, client=None, dvm_config=None):
        for tag in tags:
            if tag.as_vec()[0] == 'i':
                input_value = tag.as_vec()[1]
                input_type = tag.as_vec()[2]
                if input_type != "event" and input_type != "job" and input_type != "text":
                    return False
                # if input_type == "text" and len(input_value) > 250:
                #    return False
                if input_type == "event":
                    evt = await get_event_by_id(tag.as_vec()[1], client=client, config=dvm_config)
                    # if len(evt.content()) > 250:
                    #    return False
            elif tag.as_vec()[0] == 'param':
                param = tag.as_vec()[1]
                if param == "language":  # check for param type
                    if tag.as_vec()[2] != "en":  # todo add other available languages
                        return False

        return True

    async def create_request_from_nostr_event(self, event, client=None, dvm_config=None):
        request_form = {"jobID": event.id().to_hex() + "_" + self.NAME.replace(" ", "")}
        prompt = "test"
        if self.options.get("input_file") and self.options.get("input_file") != "":
            input_file = self.options['input_file']
        else:
            input_file = "cache/input.wav"

        language = "en"

        for tag in event.tags():
            if tag.as_vec()[0] == 'i':
                input_type = tag.as_vec()[2]
                if input_type == "event":

                    evt = await get_event_by_id(tag.as_vec()[1], client=client, config=dvm_config)

                    if evt is not None:
                        prompt = evt.content()
                    else:
                        raise FileNotFoundError("Couldn't find event")


                elif input_type == "text":
                    prompt = tag.as_vec()[1]
                elif input_type == "job":

                    evt = await get_referenced_event_by_id(event_id=tag.as_vec()[1], client=client,
                                                           kinds=[EventDefinitions.KIND_NIP90_RESULT_EXTRACT_TEXT,
                                                                  EventDefinitions.KIND_NIP90_RESULT_SUMMARIZE_TEXT,
                                                                  EventDefinitions.KIND_NIP90_RESULT_TRANSLATE_TEXT],
                                                           dvm_config=dvm_config)
                    prompt = evt.content()
                if input_type == "url":
                    input_file = tag.as_vec()[1]
            elif tag.as_vec()[0] == 'param':
                param = tag.as_vec()[1]
                if param == "language":  # check for param type
                    language = tag.as_vec()[2]
                elif param == "voice":  # check for param type
                    input_file = "cache/" + tag.as_vec()[2] + ".wav"

        if not Path.exists(Path(input_file)):
            input_file_url = "https://media.nostr.build/av/de104e3260be636533a56fd4468b905c1eb22b226143a997aa936b011122af8a.wav"
            urllib.request.urlretrieve(input_file_url, "cache/input.wav")
            input_file = "cache/input.wav"

        options = {
            "prompt": prompt,
            "input_wav": input_file,
            "language": language
        }
        request_form['options'] = json.dumps(options)

        return request_form

    async def process(self, request_form):
        import torch
        from TTS.api import TTS
        import re

        options = self.set_options(request_form)
        device = "cuda" if torch.cuda.is_available() else "cpu"
        # else "mps" if torch.backends.mps.is_available()
        print(device)

        print(TTS().list_models().list_tts_models())

        try:
            # model = "tts_models/deu/fairseq/vits"
            # model = "tts_models/multilingual/multi-dataset/your_tts"
            model = "tts_models/multilingual/multi-dataset/xtts_v2"
            tts = TTS(model).to(device)

            text = (options["prompt"].replace("~", "around ").replace("#", "hashtag")
                    .replace(" LFG", " Let's fucking go").replace("_", " ").replace("*", " ")
                    .replace("![]", " ").replace("( )", "").replace("**", " ").replace("\xa0", "")
                    )

            text_clean = re.sub(
                r'''(?i)\b((?:https?://|www\d{0,3}[.]|[a-z0-9.\-]+[.][a-z]{2,4}/)(?:[^\s()<>]+|\(([^\s()<>]+|(\([^\s()<>]+\)))*\))+(?:\(([^\s()<>]+|(\([^\s()<>]+\)))*\)|[^\s`!()\[\]{};:'".,<>?«»“”‘’]))''',
                " ", text)

            tts.tts_to_file(
                text=text_clean,
                speaker_wav=options["input_wav"], language=options["language"], file_path="outputs/output.wav")

            try:
                print("Converting Audio")
                final_filename = "outputs/output.mp3"
                fs, x = ffmpegio.audio.read("outputs/output.wav", sample_fmt='dbl', ac=1)
                ffmpegio.audio.write(final_filename, fs, x, overwrite=True)
            except:
                final_filename = "outputs/output.wav"

            result = await upload_media_to_hoster(final_filename, key_hex=self.dvm_config.PRIVATE_KEY)
            print(result)
            return result
        except Exception as e:
            print("Error in Module: " + str(e))
            raise Exception(e)


# We build an example here that we can call by either calling this file directly from the main directory,
# or by adding it to our playground. You can call the example and adjust it to your needs or redefine it in the
# playground or elsewhere
def build_example(name, identifier, admin_config):
    dvm_config = build_default_config(identifier)
    admin_config.LUD16 = dvm_config.LN_ADDRESS

    # use an alternative local wav file you want to use for cloning
    options = {'input_file': ""}

    nip89info = {
        "name": name,
        "image": "https://image.nostr.build/c33ca6fc4cc038ca4adb46fdfdfda34951656f87ee364ef59095bae1495ce669.jpg",
        "about": "I Generate Speech from Text",
        "encryptionSupported": True,
        "cashuAccepted": True,
        "nip90Params": {
            "language": {
                "required": False,
                "values": []
            }
        }
    }

    nip89config = NIP89Config()
    nip89config.DTAG = check_and_set_d_tag(identifier, name, dvm_config.PRIVATE_KEY, nip89info["image"])
    nip89config.CONTENT = json.dumps(nip89info)

    return TextToSpeech(name=name, dvm_config=dvm_config, nip89config=nip89config, admin_config=admin_config,
                        options=options)


if __name__ == '__main__':
    process_venv(TextToSpeech)



import json
import os

from nostr_sdk import Kind

from nostr_dvm.interfaces.dvmtaskinterface import DVMTaskInterface, process_venv
from nostr_dvm.utils.admin_utils import AdminConfig
from nostr_dvm.utils.dvmconfig import DVMConfig, build_default_config
from nostr_dvm.utils.nip88_utils import NIP88Config
from nostr_dvm.utils.nip89_utils import NIP89Config, check_and_set_d_tag

"""
This File contains a Generic DVM that can be overwritten by the user
Accepted Inputs: None
Outputs: Text
Params:  None
"""


class GenericDVM(DVMTaskInterface):
    KIND: Kind = Kind(5000)
    TASK: str = "generic"
    FIX_COST: float = 0
    dvm_config: DVMConfig
    options = {}

    async def init_dvm(self, name, dvm_config: DVMConfig, nip89config: NIP89Config, nip88config: NIP88Config = None,
                       admin_config: AdminConfig = None, options=None):
        dvm_config.SCRIPT = os.path.abspath(__file__)
        if dvm_config.KIND is not None:
            self.KIND = dvm_config.KIND

    async def is_input_supported(self, tags, client=None, dvm_config=None):
        return True

    async def create_request_from_nostr_event(self, event, client=None, dvm_config=None):
        self.dvm_config = dvm_config
        print(self.dvm_config.PRIVATE_KEY)
        prompt = ""
        user = event.author().to_hex()
        for tag in event.tags():
            if tag.as_vec()[0] == 'i':
                input_type = tag.as_vec()[2]
                if input_type == "text":
                    prompt = tag.as_vec()[1]
            elif tag.as_vec()[0] == 'param':
                if tag.as_vec()[1] == 'user':
                    user = tag.as_vec()[2]
                else:
                    self.options[tag.as_vec()[1]] = tag.as_vec()[2]

        request_form = {"jobID": event.id().to_hex()}

        self.options["request_event_id"] = event.id().to_hex()
        self.options["request_event_author"] = user
        if prompt != "":
            self.options["input"] = prompt
        request_form['options'] = json.dumps(self.options)
        return request_form

    async def process(self, request_form):
        options = self.set_options(request_form)
        result = "I'm manipulating the DVM from my inside function\n"
        result += options["some_option"]
        print(result)
        return result


# We build an example here that we can call by either calling this file directly from the main directory,
# or by adding it to our playground. You can call the example and adjust it to your needs or redefine it in the
# playground or elsewhere
def build_example(name, identifier, admin_config, announce=False):
    admin_config = AdminConfig()
    admin_config.REBROADCAST_NIP89 = announce
    admin_config.REBROADCAST_NIP65_RELAY_LIST = announce
    admin_config.UPDATE_PROFILE = announce

    name = "Generic DVM"
    identifier = "a_very_generic_dvm"  # Chose a unique identifier in order to get a lnaddress
    dvm_config = build_default_config(identifier)
    dvm_config.KIND = Kind(5050)  # Manually set the Kind Number (see data-vending-machines.org)

    # Add NIP89
    nip89info = {
        "name": name,
        "image": "https://image.nostr.build/28da676a19841dcfa7dcf7124be6816842d14b84f6046462d2a3f1268fe58d03.png",
        "about": "I'm an all purpose DVM'",
        "encryptionSupported": True,
        "cashuAccepted": True,
        "nip90Params": {
        }
    }

    nip89config = NIP89Config()
    nip89config.DTAG = check_and_set_d_tag(identifier, name, dvm_config.PRIVATE_KEY, nip89info["image"])
    nip89config.CONTENT = json.dumps(nip89info)

    options = {
        "some_option": "#RunDVM",
    }

    dvm = GenericDVM(name=name, dvm_config=dvm_config, nip89config=nip89config,
                     admin_config=admin_config, options=options)

    async def process(request_form):
        options = dvm.set_options(request_form)
        result = "I'm manipulating the DVM from outside\n"
        result += options["some_option"]
        print(result)
        return result

    dvm.process = process  # overwrite the process function with the above one
    return dvm


if __name__ == '__main__':
    process_venv(GenericDVM)



import json
import os

from nostr_sdk import Kind

from nostr_dvm.interfaces.dvmtaskinterface import DVMTaskInterface, process_venv
from nostr_dvm.utils.admin_utils import AdminConfig
from nostr_dvm.utils.definitions import EventDefinitions
from nostr_dvm.utils.dvmconfig import DVMConfig, build_default_config
from nostr_dvm.utils.mediasource_utils import organize_input_media_data
from nostr_dvm.utils.nip88_utils import NIP88Config
from nostr_dvm.utils.nip89_utils import NIP89Config
from nostr_dvm.utils.output_utils import upload_media_to_hoster

"""
This File contains a Module convert media locally

Accepted Inputs: Text, Events, Jobs (Text Extraction, Summary, Translation)
Outputs: Text containing the TranslationGoogle in the desired language.
Params:  -language The target language
"""


class MediaConverter(DVMTaskInterface):
    KIND: Kind = EventDefinitions.KIND_NIP90_CONVERT_VIDEO
    TASK = "convert"
    FIX_COST = 20
    PER_UNIT_COST = 0.1

    async def init_dvm(self, name, dvm_config: DVMConfig, nip89config: NIP89Config, nip88config: NIP88Config = None,
                       admin_config: AdminConfig = None, options=None):
        dvm_config.SCRIPT = os.path.abspath(__file__)

    async def is_input_supported(self, tags, client=None, dvm_config=None):
        for tag in tags:
            if tag.as_vec()[0] == 'i':
                input_value = tag.as_vec()[1]
                input_type = tag.as_vec()[2]
                if input_type != "url":
                    return False
        return True

    async def create_request_from_nostr_event(self, event, client=None, dvm_config=None):
        request_form = {"jobID": event.id().to_hex()}
        url = ""
        media_format = "video/mp4"
        input_type = "text"
        start_time = 0
        end_time = 0
        # TODO parse start/end parameters

        for tag in event.tags():
            if tag.as_vec()[0] == 'i':
                input_type = tag.as_vec()[2]
                if input_type == "url":
                    url = str(tag.as_vec()[1]).split('#')[0]


            elif tag.as_vec()[0] == 'param':
                param = tag.as_vec()[1]
                if param == "format":  # check for param type
                    media_format = tag.as_vec()[2]

        filepath = await organize_input_media_data(url, input_type, start_time, end_time, dvm_config, client, True,
                                                   media_format)
        options = {
            "filepath": filepath
        }

        request_form['options'] = json.dumps(options)
        return request_form

    async def process(self, request_form):
        options = self.set_options(request_form)
        url = await upload_media_to_hoster(options["filepath"])

        return url


# We build an example here that we can call by either calling this file directly from the main directory,
# or by adding it to our playground. You can call the example and adjust it to your needs or redefine it in the
# playground or elsewhere
def build_example(name, identifier, admin_config):
    dvm_config = build_default_config(identifier)
    dvm_config.USE_OWN_VENV = False
    admin_config.LUD16 = dvm_config.LN_ADDRESS
    # Add NIP89
    nip89info = {
        "name": name,
        "image": "https://image.nostr.build/c33ca6fc4cc038ca4adb46fdfdfda34951656f87ee364ef59095bae1495ce669.jpg",
        "about": "I convert videos from urls to given output format.",
        "encryptionSupported": True,
        "cashuAccepted": True,
        "nip90Params": {
            "format": {
                "required": False,
                "values": ["video/mp4", "audio/mp3"]
            }
        }
    }

    nip89config = NIP89Config()

    return MediaConverter(name=name, dvm_config=dvm_config, nip89config=nip89config,
                          admin_config=admin_config)


if __name__ == '__main__':
    process_venv(MediaConverter)



import json
import os

import requests
from nostr_sdk import Keys, PublicKey, NostrWalletConnectUri, Nwc

from nostr_dvm.utils.dvmconfig import DVMConfig
from nostr_dvm.utils.nostr_utils import check_and_set_private_key
from nostr_dvm.utils.zap_utils import zaprequest


async def nwc_zap(connectionstr, bolt11, keys, externalrelay=None):
    uri = NostrWalletConnectUri.parse(connectionstr)

    # Initialize NWC client
    nwc = Nwc(uri)

    info = await nwc.get_info()
    print(info)

    balance = await nwc.get_balance()
    print(f"Balance: {balance} MilliSats")

    event_id = await nwc.pay_invoice(bolt11)
    print("NWC event: " + event_id)
    return event_id


def parse_connection_str(connectionstring):
    split = connectionstring.split("?")
    targetpubkey = split[0].split(":")[1].replace("//", "")
    split2 = split[1].split("&")
    relay = split2[0].split("=")[1]
    relay = relay.replace("%3A%2F%2F", "://")
    secret = split2[1].split("=")[1]
    return targetpubkey, relay, secret


def make_nwc_account(identifier, nwcdomain):
    pubkey = Keys.parse(os.getenv("DVM_PRIVATE_KEY_" + identifier.upper())).public_key().to_hex()
    data = {
        'name': identifier,
        'host': os.getenv("LNBITS_HOST"),
        'key': os.getenv("LNBITS_ADMIN_KEY_" + identifier.upper()),
        'pubkey': pubkey,
    }

    try:
        url = nwcdomain
        header = {"content-type": "application/json"}
        res = requests.post(url, headers=header, json=data)
        obj = json.loads(res.text)
        if obj['params']['connectionURI'] != "Users already exists":
            return obj['params']['connectionURI']
        else:
            return ""

    except Exception as e:
        print(e)
        return ""


async def nwc_test(nwc_server):
    connectionstring = make_nwc_account("test", nwc_server + "/api/new")
    print(connectionstring)
    #  TODO Store the connection string in a db, use here if you already have one
    # connectionstring = "nostr+walletconnect:..."
    if connectionstring != "":
        # we use the keys from a test user
        keys = Keys.parse(check_and_set_private_key("test"))

        # we zap npub1nxa4tywfz9nqp7z9zp7nr7d4nchhclsf58lcqt5y782rmf2hefjquaa6q8's profile 21 sats and say Cool stuff
        pubkey = PublicKey.from_bech32("npub1nxa4tywfz9nqp7z9zp7nr7d4nchhclsf58lcqt5y782rmf2hefjquaa6q8")
        bolt11 = zaprequest("hype@bitcoinfixesthis.org", 21, "Cool Stuff", None,
                            pubkey, keys, DVMConfig.RELAY_LIST)

        await nwc_zap(connectionstring, bolt11, keys)
